/***************************************************************************
                chessterfielden.cpp  -  ChessterfieldEN main
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// disable warning C4786: symbol greater than 255 character,
#ifndef GNU_VERSION
#pragma warning(disable: 4786) // Visual C++ stuff
#endif

#ifndef GNU_VERSION
#include <conio.h>
#include <process.h>
#endif

#include "def_inc.h"
// #include "mytcp.h"
#include "myvector.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "evaluation.h"
#include "pgnfile.h"
#include "learnobject.h"

#include <iostream>
#include <string>
#include <ctime>
#include <vector>
#include <fstream>

using namespace std;

typedef vector<string> STRINGVECTOR;

int main()
{
	clock_t start, stop;

	cout << endl;
	cout << "---------------------------------------------------------\n";
	cout << "          ChessterfieldEN - Learning Utility\n";
	cout << "---------------------------------------------------------\n";
	cout << "Author: Matthias Luescher (m.luescher@datacomm.ch)\n";
	cout << "Copyright: Matthias Luescher\n";
	cout << "Web: http://www.luescher-online.com/computerchess.html\n";
	cout << "Version: i5a (unofficial release)\n\n";
	cout << "Tip: Consult the file readme.txt for a short introduction\n";
	cout << "     to the ChessterfieldEN learning utility.\n";
	cout << "---------------------------------------------------------\n\n";

	LearnObject lrn;	
	string str;

	while ( str != "quit" )
	{
		cout << "Enter your command ('help' shows available commands): ";
		cin >> str;
		cout << "\n";

		if ( str == "help" )
		{
			cout << "\n";
			cout << "The following commands are available: \n\n";
			cout << "pgn       learn from user supplied chess games (pgn-format)\n";
			cout << "optimize  optimize the network.dat file\n";
			cout << "sdb       learn from (proprietary) endgame databases (.sdb files)\n";
			cout << "stat      do some statistics on endgame databases (.sdb files)\n";
			cout << "auto      use WinBoard to play games against other engines and\n";
			cout << "          learn from these games \n";
			cout << "          (see /chessterfield/data/config.dat for more information)\n";
			cout << "quit      leave the program\n\n";
		}
		else if ( str == "optimize" )
		{
			cout << "Optimizing configuration set ... ";
			lrn.OptimizeConfigurationSet();
			cout << "done.\n\n";
		}
		else if ( str == "sdb" )
		{
			int filter = 0;
			cout << "Enter the filename of the training set: ";
			cin >> str;
			cout << "Position filter number ('0' for no filter): ";
			cin >> filter;
			cout << "\n";
			if (!cin.good()) cin.clear();
			start = clock();
			if ( lrn.LoadTrainingSet(str, filter) ) 
			{
				lrn.GenerateConfigurations(3);
				lrn.WeightFitting();
				lrn.WDLStatistics();
			}
			stop = clock();
			double duration = (double)(stop - start) / CLOCKS_PER_SEC;
			cout << "Time: " << duration << " seconds\n\n";
		}
		else if ( str == "stat" )
		{
			int filter = 0;
			// int sampling = 1;
			cout << "Enter the filename of the training set: ";
			cin >> str;
			cout << "Position filter number ('0' for no filter): ";
			cin >> filter;
			if (!cin.good()) cin.clear();
			// cout << "Sampling step size (must be bigger than zero): ";
			// cin >> sampling;
			// if (!cin.good()) cin.clear();
			cout << "\n";

			// if ( sampling < 1 ) sampling = 1;
			
			if ( lrn.LoadTrainingSet(str, filter) ) 
			{
				lrn.WDLStatistics();
				// lrn.HalfmoveStatistics(str, filter, sampling);
			}
		}
		else if ( str == "pgn" )
		{
			PGNFile f;
			cout << "Enter the name of a pgn file or type 'quit' to exit: ";
			cin >> str;
			
			if ( str != "quit" )
			{
				do
				{	
					f.LoadFile(str);
					cout << "Load another file? ('somefile.pgn' or 'no'): ";
					cin >> str;
				} while ( str != "no" );

				cout << "\n";
				start = clock();

				lrn.ProcessPGNFile(&f);
				lrn.GeneratePGNConfigurations();
				lrn.PGNWeightFitting();
				stop = clock();
				double duration = (double)(stop - start) / CLOCKS_PER_SEC;
				cout << "Time: " << duration << " seconds\n\n";
			}	
		}
		else if ( str == "auto" )
		{
#ifdef GNU_VERSION
			cout << "Sorry, this command is not implemented in the unix version."
				 << endl;
#else
			cout << "NOTE: The program will now read the /data/config.dat file.\n"
				 << "      Please have a look at this file and do some\n" 
				 << "      reconfiguration where needed.\n\n"
				 << "Do you want to continue? ('yes', 'no'): ";
			cin >> str;
			cout << endl;

			string winboarddir, executable, pgnfile, winboardoptions;
			string firstdir, firstengine;
			STRINGVECTOR secondengines;

			// parsing the config.dat file ->
			fstream infile;
			infile.open("config.dat", ios::in);
			if ( infile.good() && str == "yes")
			{
				infile >> str;
				while ( infile.good() )
				{
					if ( str[0] == '#' ) getline(infile, str, '\n');
					else if ( str == "WINBOARDDIR" ) 
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						winboarddir = str;
					}
					else if ( str == "EXECUTABLE" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						executable = str;
					}
					else if ( str == "PGNFILE" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						pgnfile = str;
					}
					else if ( str == "WINBOARDOPTIONS" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						winboardoptions = str;
					}
					else if ( str == "FIRSTDIR" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						firstdir = str;
					}
					else if ( str == "FIRSTENGINE" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						firstengine = str;
					}
					else if ( str == "SECONDDIR" )
					{
						getline(infile, str, '[');
						getline(infile, str, ']');
						secondengines.push_back(str);
						infile >> str;
						if ( str == "SECONDENGINE" )
						{
							getline(infile, str, '[');
							getline(infile, str, ']');
							secondengines.push_back(str);
						}
						else
						{
							cout << "Error: Missing SECONDENGINE tag in config.dat file!" 
								 << endl;
						}
					}

					infile >> str;
				}
				infile.close();
				// <- parsing the config.dat file

				cout << "Hit any key to interrupt the program!\n";
				cout << "REMARK: Wait until the program has properly finished and\n" 
					 << "do NOT press 'ctrl-c'!\n";
				do
				{
					int i = 0;
					while ( i + 1 < secondengines.size() && !kbhit() )
					{
						string t1, t2;
						t1 = winboarddir + "\\" + executable;
						t2 = "/saveGameFile=\"" + pgnfile + "\"";
						_spawnl( _P_WAIT, t1.c_str(), executable.c_str(),  
							     t2.c_str(), winboardoptions.c_str(),
								 "/cp",
								 firstdir.c_str(), firstengine.c_str(),
								 secondengines[i].c_str(), 
								 secondengines[i+1].c_str(), NULL);
						i += 2;
					}

					if ( !kbhit() )
					{
						PGNFile f;
						f.LoadFile(pgnfile);
						start = clock();
						lrn.ProcessPGNFile(&f);
						lrn.GeneratePGNConfigurations();
						lrn.PGNWeightFitting();
						stop = clock();
						double duration = (double)(stop - start) / CLOCKS_PER_SEC;
						cout << "Time: " << duration << " seconds\n\n";
					}
				} while ( !kbhit() ); 
			}
			else if ( str == "yes" )
			{
				cout << "Error: config.dat file not found!\n";
			}
#endif
		}
	}
	
	return 0;
}
