/***************************************************************************
                      def_inc.h  -  define statements
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string>

using namespace std;

// Debug Support
#ifdef NDEBUG
#define DEBUG_OUTPUT ((void)0)
#else
#define DEBUG_OUTPUT printf
#endif

class Move 
{ 
public:
	int sq1, sq2, special, remark;
	int value, o_value;
	bool check_test, toPlay, valid;
};

class Piece;

class ExtendedMove
{ 
public:
	Piece* pP1; 
	Piece* pP2;
	Piece* pP3;
	int sq1, sq2, sq3, sq4;
	int special;
	int epSquare;
};

// engine search modes
#define TIME_LIMIT 0
#define DEPTH_LIMIT 1
#define NO_LIMITS 2

// values for HashTable::m_flags
#define NO_TYPE 1
#define IN_WINDOW 2
#define FAIL_HIGH 4
#define FAIL_LOW 8
#define TYPE_MASK 15
#define SIDE_TO_MOVE 16

// values for Move::special
#define NOTHING 0
#define QUEEN_CASTLING 1
#define KING_CASTLING 2
#define EN_PASSANT 3
#define NEW_QUEEN 4
#define NEW_KNIGHT 5
#define NEW_BISHOP 6
#define NEW_ROOK 7
#define FIRST_MOVE 10
#define IN_LINE 11
#define DOUBLE_ADVANCE 12 // pawn advances two squares -> en passant square

// values for Move::remark
#define NEW_MOVE 0
#define DO_MOVE 1 

#define MOVECOUNT_LIMIT 2 // currently, the number of moves is limited :-(
#define VERIFY_MOVE 3  // castling? en passant? check? valid move?

#define NEW_GAME 5
#define CHECKMATE 7
#define PATT 9  
#define REMIS 10
#define REPETITION_RULE 11


// values for Move::toPlay:
//      BLACK 0 // DONT CHANGE THIS VALUE
//      WHITE 1 // DONT CHANGE THIS VALUE

// concerns "piece 32": (all empty files are occupied by this piece)
#define NO_PIECE_NR 32 // DONT CHANGE THIS VALUE
#define NO_COLOUR 2 // DONT CHANGE THIS VALUE
#define NO_SQUARE 64 // DONT CHANGE THIS VALUE
#define NO_PIECE 6 // DONT CHANGE THIS VALUE

#define IMPOSSIBLE_VALUE -300000L 
#define CHECKMATE_VALUE 200000L
#define ENDGAME_DRAW_VALUE 10
#define ENDGAME_VALUE_OFFSET 5

#define MAX_NR_OF_MOVES 400

// pieces:
#define PAWN 0 // DONT CHANGE THIS VALUE
#define KNIGHT 1 // DONT CHANGE THIS VALUE
#define BISHOP 2 // DONT CHANGE THIS VALUE
#define ROOK 3 // DONT CHANGE THIS VALUE
#define QUEEN 4 // DONT CHANGE THIS VALUE
#define KING 5 // DONT CHANGE THIS VALUE
//		NO_PIECE 6 // DONT CHANGE THIS VALUE

// piece colors:
#define BLACK 0 // DONT CHANGE THIS VALUE
#define WHITE 1 // DONT CHANGE THIS VALUE
//		NO_COLOUR 2 // DONT CHANGE THIS VALUE

// pieces for initial board setup (DONT CHANGE THIS VALUES)
#define B_PAWN 0 // these values should correspond to PGN_..._...
#define B_ROOK 1
#define B_BISHOP 2
#define B_KNIGHT 3
#define B_QUEEN 4
#define B_KING 5
#define W_PAWN 6
#define W_ROOK 7
#define W_BISHOP 8
#define W_KNIGHT 9
#define W_QUEEN 10
#define W_KING 11
#define EMPTY_SQUARE 12

#define MAX_SEARCH_DEPTH 40 // maximal search depth
							// can be changed but MAX_SEARCH_DEPTH MUST be an even number

#define DRAW_VALUE_PAWN 5
#define DRAW_VALUE_KNIGHT 2
#define DRAW_VALUE_BISHOP 3
#define DRAW_VALUE_ROOK 5
#define DRAW_VALUE_QUEEN 5
#define DRAW_VALUE_KING 0
#define DRAW_VALUE_NO_PIECE 0
#define CRITICAL_DRAW_VALUE 5 // positions with draw values below five are draw 
							  // NOTE: there are special endgames where this calculation
							  // is incorrect


#define PVAR_BONUS 30000
#define HASH_BONUS 25000
#define KILLER1BONUS 20000
#define KILLER2BONUS 15000


#define BIT_0	(unsigned int) 1U
#define BIT_1	(unsigned int) 2U
#define BIT_2	(unsigned int) 4U

#define BIT_3	(unsigned int) 8U
#define BIT_4	(unsigned int) 16U
#define BIT_5	(unsigned int) 32U
#define BIT_6	(unsigned int) 64U
#define BIT_7	(unsigned int) 128U
#define BIT_8	(unsigned int) 256U
#define BIT_9	(unsigned int) 512U
#define BIT_10	(unsigned int) 1024U
#define BIT_11	(unsigned int) 2048U
#define BIT_12	(unsigned int) 4096U
#define BIT_13	(unsigned int) 8192U
#define BIT_14	(unsigned int) 16384U
#define BIT_15	(unsigned int) 32768U
#define BIT_16	(unsigned int) 65536U
#define BIT_17	(unsigned int) 131072U
#define BIT_18	(unsigned int) 262144U
#define BIT_19	(unsigned int) 524288U
#define BIT_20	(unsigned int) 1048576U
#define BIT_21	(unsigned int) 2097152U
#define BIT_22	(unsigned int) 4194304U
#define BIT_23	(unsigned int) 8388608U
#define BIT_24	(unsigned int) 16777216U
#define BIT_25	(unsigned int) 33554432U
#define BIT_26	(unsigned int) 67108864U
#define BIT_27	(unsigned int) 134217728U
#define BIT_28	(unsigned int) 268435456U
#define BIT_29	(unsigned int) 536870912U
#define BIT_30	(unsigned int) 1073741824U
#define BIT_31	(unsigned int) 2147483648U

#define Q_QUEEN  9000
#define Q_KING      0
#define Q_ROOK   5000
#define Q_BISHOP 3000
#define Q_KNIGHT 3000
#define Q_PAWN   1000

#define Q_EVALUATION_WINDOW 600

// network.dat file version
#define FILE_VERSION 1

// position types
#define KRK					0		
#define KPK					1
#define KQK					2
#define KBBK				3
#define KPXKPX				4 // Number of Pawns > 1 && Number of Kings == 2 && No other pieces
#define KBNK				5

#define ENDGAME_1			6  
#define Q_ENDGAME_1		10000 // quality count < 10 * Q_PAWN

#define ENDGAME_2			7  
#define Q_ENDGAME_2     20000

#define MIDGAME_1			8
#define Q_MIDGAME_1		30000

#define MIDGAME_2			9
#define Q_MIDGAME_2		40000

#define MIDGAME_3		   10
#define Q_MIDGAME_3		50000

#define MIDGAME_4		   11
#define Q_MIDGAME_4     60000

#define MIDGAME_5		   12
#define Q_MIDGAME_5		70000

#define OPENING			   13
//      Q_OPENING		xxxxx // quality count < inf


#define MAX_POSITIONTYPES  14 

#define	AUTO_CHOOSE_POSITION_TYPE -1
