/***************************************************************************
                 engine.cpp  -  tree search for ChessterfieldCL
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GNU_VERSION
#include <windows.h>
#endif

#include "def_inc.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "hashtable.h"
#include "evaluation.h"

#include "engine.h"

#include <cassert>
#include <iostream>
#include <string>
#include <ctime>

using namespace std;

// Engine
//
/////////////////////////////////////////////////////////////////////////////

Engine::Engine(int* board, bool toPlay)
{
int defaultpos[] = { 
	W_ROOK, W_KNIGHT, W_BISHOP, W_QUEEN, W_KING, W_BISHOP, W_KNIGHT, W_ROOK,
	W_PAWN, W_PAWN,   W_PAWN,   W_PAWN,  W_PAWN, W_PAWN,   W_PAWN,   W_PAWN,
	EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE,
	EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE,
	EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE,
	EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE, EMPTY_SQUARE,
	B_PAWN, B_PAWN,   B_PAWN,   B_PAWN,  B_PAWN, B_PAWN,   B_PAWN,   B_PAWN,
	B_ROOK, B_KNIGHT, B_BISHOP, B_QUEEN, B_KING, B_BISHOP, B_KNIGHT, B_ROOK};

	if ( board == NULL ) board = defaultpos;

	m_pv1Move.valid = false;

	m_move.toPlay = toPlay;

	m_bPostThinking = false;
	m_bAnalyzeMode = false;
	m_zeroMoveCount = 0;
	m_move.remark = NEW_GAME;

	m_bStopEngine = false;
	m_bSoftStop = false;
	m_bPostAnalysis = false;

	m_bRequestHash = true;
	m_bPostThinking = true;

    // place all pieces on the board
	int pieceNr, square, currentHashtype;
    int hashtype2type[] = { PAWN, ROOK, BISHOP, KNIGHT, QUEEN, KING,
					        PAWN, ROOK, BISHOP, KNIGHT, QUEEN, KING };

	m_move.special = FIRST_MOVE;
	m_move.remark = NEW_MOVE;

	for ( square = 0; square < 65; square++ ) m_board[square] = NULL;

	currentHashtype = W_KING;
	pieceNr = 15;

	new Piece(NO_COLOUR, NO_SQUARE, NO_PIECE_NR, NULL );
	while ( pieceNr >= 0 )
	{
		for ( square = 0; square < 64; square++ )
		{
			if ( board[square] == currentHashtype )
			{
				switch ( hashtype2type[currentHashtype] )
				{
				case PAWN: new WPawn(WHITE, square, pieceNr, NULL); break;
				case KNIGHT: new Knight(WHITE, square, pieceNr, NULL); break;
				case BISHOP: new Bishop(WHITE, square, pieceNr, NULL); break;
				case ROOK: new Rook(WHITE, square, pieceNr, NULL); break;
				case QUEEN: new Queen(WHITE, square, pieceNr, NULL); break;
				case KING: m_pKing[WHITE] = new King(WHITE, square, pieceNr, NULL); break;
				}
				pieceNr--;
			}
		}
		currentHashtype--;

		if ( currentHashtype < 6 )
		{
			for ( pieceNr = pieceNr; pieceNr >= 0; pieceNr-- )
				new WPawn(WHITE, NO_SQUARE, pieceNr, NULL);
		}
	}

	currentHashtype = B_KING;
	pieceNr = 31;
	while ( pieceNr >= 16 )
	{
		for ( square = 0; square < 64; square++ )
		{
			if ( board[square] == currentHashtype )
			{
				switch ( hashtype2type[currentHashtype] )
				{
				case PAWN: new BPawn(BLACK, square, pieceNr, NULL); break;
				case KNIGHT: new Knight(BLACK, square, pieceNr, NULL); break;
				case BISHOP: new Bishop(BLACK, square, pieceNr, NULL); break;
				case ROOK: new Rook(BLACK, square, pieceNr, NULL); break;
				case QUEEN: new Queen(BLACK, square, pieceNr, NULL); break;
				case KING: m_pKing[BLACK] = new King(BLACK, square, pieceNr, NULL); break;
				}
				pieceNr--;
			}
		}
		currentHashtype--;

		if ( currentHashtype < 0 )
		{
			for ( pieceNr = pieceNr; pieceNr >= 16; pieceNr-- )
				new BPawn(BLACK, NO_SQUARE, pieceNr, NULL);
		}
	}

	for ( square = 0; square < 65; square++ ) 
	{
		if ( m_board[square] == NULL )
			m_board[square] = m_pieceList[NO_PIECE_NR];
	}

	if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
	{
		m_move.remark = CHECKMATE;
		m_move.valid = false;
	}

	// initialize hash table
	m_hash->Initialize();
	m_hash->IncrementConfigurationCount();
}

Engine::~Engine()
{
	DeallocateStaticMembers();
}

string Engine::MoveToString(int sq1, int sq2, int special)
{ 
	string s = "";

	if ( sq1 >= 0 && sq1 < 64 && 
		 sq2 >= 0 && sq2 < 64)
	{
		s += (sq1 & 7) + 'a';
		s += (sq1 >> 3) + '1';
		s += (sq2 & 7) + 'a';
		s += (sq2 >> 3) + '1';
	}
	else
		return s;

	switch( special )
	{
	case NEW_QUEEN: s += 'q'; break;
	case NEW_ROOK: s += 'r'; break;
	case NEW_BISHOP: s += 'b'; break;
	case NEW_KNIGHT: s += 'n'; break;
	}

	return s;
}

bool Engine::MakeComputerMove(long tSoftStop, long tHardStop, int searchDepth, 
							  int searchMode, bool bResetHash) 
{
	m_timeSoftStop = tSoftStop;
	m_timeHardStop = tHardStop;
	m_searchMode = searchMode;

	if ( searchMode != DEPTH_LIMIT )
		searchDepth = MAX_SEARCH_DEPTH;

	if ( m_gameHistory->GetCount() + 2 >= MAX_NR_OF_MOVES )
	{
		m_move.remark = MOVECOUNT_LIMIT;
		return false;
	}

    if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
	{
		m_move.remark = CHECKMATE;
		m_move.valid = false;
		return false;
	}

	if ( m_hash->GetConfigurationCount() > 2 )
	{
		m_move.remark = REPETITION_RULE;
		m_move.valid = false;
		return false;
	}

	m_pv1Move.valid = false;

	m_nodeCount = 0;
	m_timeOffset = GetTickCount();

NodeStack stackL1(m_moveStack); 
Move* best;

	if ( searchDepth > MAX_SEARCH_DEPTH / 2 - 1 )
		searchDepth = MAX_SEARCH_DEPTH / 2 - 1;

	// earlier versions of Chessterfield had multiple move generators fighting for 
	// only one hash table, nowadays there is only one move generator and no more fight 
	m_hash->RequestHash(m_bRequestHash, bResetHash);
	root_alphabeta(stackL1, searchDepth);
	m_hash->ReleaseHash();

	stackL1.BestOrientation();


	if ( m_bAnalyzeMode )
	{
		m_move.remark = NEW_MOVE;
		m_move.valid = false;
		return false;
	}

	if ( (best = stackL1.GetNextBest()) != NULL && best->valid )
	{
		m_move = *best;
		DEBUG_OUTPUT("\nNode count: %i...Value: %i", m_nodeCount, m_move.value); 

		pushMove();
		m_move.remark = NEW_MOVE;

		Move m = m_move;
		m_move.sq1 = m_PV[0][1][0];
		m_move.sq2 = m_PV[0][1][1];
		m_move.special = NOTHING;
		IsValidMove();
		m_pv1Move = m_move;
		m_move = m;

		if ( m_move.value == CHECKMATE_VALUE - 1 )
			m_move.remark = CHECKMATE;

		if ( m_hash->GetConfigurationCount() > 2 )
			m_move.remark = REPETITION_RULE;

		m_move.check_test = m_pKing[m_move.toPlay]->isAttacked();

		return true;
	}
	else
	{
		m_move.valid = false;
		if ( m_pKing[m_move.toPlay]->isAttacked() )
			m_move.remark = CHECKMATE;
		else
			m_move.remark = PATT;
		return false;
	}

	return false;
}

bool Engine::UndoMove()
{
	m_pv1Move.valid = false;

	if ( m_gameHistory->GetCount() > 0 )
	{
		popMove();
		m_move.remark = NEW_MOVE;
		return true;
	}
	else
	{						
		m_move.remark = NEW_MOVE;
		if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
		{
			m_move.remark = CHECKMATE;
			m_move.valid = false;
		}
		return false;
	}
}
	
bool Engine::MakeUserMove(int from, int to, int castlingType)
{
	if ( m_gameHistory->GetCount() + 2 >= MAX_NR_OF_MOVES )
	{
		m_move.remark = MOVECOUNT_LIMIT;
		return false;
	}

	m_pv1Move.valid = false;

	m_move.sq1 = from;
	m_move.sq2 = to;
	m_move.special = castlingType;

	if ( IsValidMove() )
	{
		pushMove();
		m_move.remark = NEW_MOVE;
		m_move.check_test = false;
		m_move.valid = true;

		if ( m_hash->GetConfigurationCount() > 2 )
			m_move.remark = REPETITION_RULE;

		return true;
	}
	return false;
}

bool Engine::generateAllMoves(NodeStack &stack, bool check, int depth)
{
Piece* p; 
int checkfreeMoves = 0, checkfreePieces = 0;
int count, nMoves, hashSq1, hashSq2;

stack.Clear();
Move* m = stack.GetTail();
	m_currentStack = &stack;

int hvarSq1 = m_PV[0][depth][0];
int hvarSq2 = m_PV[0][depth][1];
int killer1Sq1 = m_killer[depth][0][0];
int killer1Sq2 = m_killer[depth][0][1];
int killer2Sq1 = m_killer[depth][1][0];
int killer2Sq2 = m_killer[depth][1][1];
	m_hash->GetMove(hashSq1, hashSq2);

	m_move.value = -CHECKMATE_VALUE;
	m_move.o_value = 0;
	p = m_pKing[m_move.toPlay]->next;

	do
	{
		if ( check || p->type == KING || p->onKingsLine() )
		{
			// Important: Special treatement for the en passant move
			m_move.check_test = true;
		}
		else
		{
			// Important: Special treatement for the en passant move
			// check_test could be false instead of true after an en passant move
			// because the pawn on the en passant square cold be on a kings line
			m_move.check_test = false;
			checkfreePieces++;
		}

		nMoves = p->generateMoves(); 

		if ( !m_move.check_test ) 
		{
			checkfreeMoves += nMoves;
		}

		for ( count = 0; count < nMoves; count++, m++ )
		{
			if ( hvarSq1 == m->sq1 && hvarSq2 == m->sq2 )
				m->o_value += PVAR_BONUS;
			else if ( hashSq1 == m->sq1 && hashSq2 == m->sq2 )
			{
				m->o_value += HASH_BONUS;
			}
			else if ( killer1Sq1 == m->sq1 && killer1Sq2 == m->sq2 )
				m->o_value += KILLER1BONUS;
			else if ( killer2Sq1 == m->sq1 && killer2Sq2 == m->sq2 )
				m->o_value += KILLER2BONUS;

			if ( m_board[m->sq2]->color != NO_COLOUR || m->special == EN_PASSANT )
			{
				m->o_value += 10000 - (p->value - //test
										 m_board[m->sq2]->value)
										 + p->bonus[m->sq2]; // piece bonus
			}
			else
			{
				m->o_value += p->bonus[m->sq2]; // piece bonus
			}
		}

		p = p->next;
	} while ( p->previous->type != KING );

	if ( checkfreePieces < 3 || checkfreeMoves < 6 ) 
	{
		return false;
	}

	return true;
}

void Engine::generateAllXMoves(NodeStack &stack)
{
Piece* p; 
stack.Clear();

	m_currentStack = &stack;
	m_move.value = -CHECKMATE_VALUE;

	p = m_pKing[m_move.toPlay];
	do
	{
		p->generateXMoves(); 
		p = p->next;
	} while ( p->type != KING );
}

void Engine::root_alphabeta(NodeStack &stackL1, int searchDepth)
{
	m_nodeCount++; 
	for ( int s = 0; s < MAX_SEARCH_DEPTH; s++ )
		m_PV[0][s][0] = NO_SQUARE; // initialize principle variant

int value = 0, attackvalue = 0;
Move* m; 
bool checkmate = true, check = false, first = true, softStop = false, faillow = false;
int alpha, beta, branchValue = 0;

	InitializeEvaluation(branchValue);
	check = m_pKing[m_move.toPlay]->isAttacked();
bool zeroMove2 = generateAllMoves(stackL1, check, 0);

NodeStack stackL2(stackL1.GetTail()); 
	m_analysisMvTot = 0;

//--- Iterative Deepening ------------------------------------------------------//

	// possible improvements: 
	//		- verify checkmate
	//		- if there is only one possible move, no more search is required
	for ( int currentDepth = 2; 
		  ( currentDepth <= searchDepth && !m_bStopEngine && !softStop ) ||
		  currentDepth == 2; // search each move at least once 
		  currentDepth++ )  
	{
		m_baseDepth = currentDepth;

		if ( currentDepth == 2 )
		{
			alpha = IMPOSSIBLE_VALUE; 
			beta = -IMPOSSIBLE_VALUE; 
		}
		else
		{
			beta = alpha + 1100;
			alpha = alpha - 1100;
		}

		first = true;
		faillow = false;

		DEBUG_OUTPUT("\nCurrent search depth: %i", currentDepth);
		m_analysisMvDone = 0;

//--- Try all Moves ------------------------------------------------------------//
		while ( (m = stackL1.GetNextBest()) != NULL 
			     && alpha != (CHECKMATE_VALUE - 1) ) // -> best move has been found
		{						
			m_move = *m;
			attackvalue = pushMove();

			if ( m_bSoftStop && !faillow ) softStop = true;

//--- Detect Illegal Moves -----------------------------------------------------//
			if ( m->check_test && ( !m->valid ||
				 m_pKing[!m_move.toPlay]->isAttacked() ) )
			{
				m->valid = false;
				m->value = IMPOSSIBLE_VALUE;
				popMove();
			}
			else
			{
				m_analysisMvDone++;
				if ( m_baseDepth == 2 )
					m_analysisMvTot++; // count number of possible moves

//--- Do Standard Alpha-Beta ---------------------------------------------------//
				checkmate = false; 
				m->check_test = false;

				if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2 )
				{
					value = core_alphabeta( stackL2, -beta, -alpha, 1, currentDepth, 
							   -(branchValue + attackvalue), zeroMove2 ); 
				}

				if ( first )
				{
					DEBUG_OUTPUT("First move is from %i to %i\n",m->sq1, m->sq2);
					if ( value < alpha )
					{
//--- First Move Produced a Faillow --------------------------------------------//
						if ( m_bAnalyzeMode )
							cout << "--" << endl;

						faillow = true;
						beta = value;
						if ( beta <= - CHECKMATE_VALUE + MAX_SEARCH_DEPTH )
							beta++; // test! //should be more efficient when detecting
									// checkmate
						alpha = IMPOSSIBLE_VALUE;

						DEBUG_OUTPUT("\n Faillow");
						DEBUG_OUTPUT(" WINDOW: %i, %i ",alpha,beta);

						if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
						{	
							value = core_alphabeta( stackL2, -beta, -alpha, 1, currentDepth, 
								   -(branchValue + attackvalue), zeroMove2 ); 
						}
						DEBUG_OUTPUT("\tValue faillow 1: %i", value);
					}
					else if ( value >= beta ) //test: is the new move always better??
					{
//--- First Move Produced a Failhigh -------------------------------------------//
						if ( m_bAnalyzeMode )
							cout << "++" << endl;

						alpha = value;
						if ( alpha >= CHECKMATE_VALUE - MAX_SEARCH_DEPTH )
							alpha--; // test! //should be more efficient when detecting checkmate
						beta = -IMPOSSIBLE_VALUE;
						DEBUG_OUTPUT("\n Failhigh");

						if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
						{	
							value = core_alphabeta( stackL2, -beta, -alpha, 1, currentDepth, 
								   -(branchValue + attackvalue), zeroMove2 ); 
						}
						DEBUG_OUTPUT("\tValue failhigh 1: %i", value);
					}
					alpha = value;
					beta = alpha + 1;

					if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
					{ // dont do an update, if the engine has been stopped
						extendPVAR(0, m);
						postThinking(value);
						m->value = value; 
					}
				}
				else
				{
//--- Check Remaining Moves ----------------------------------------------------//
					if ( value >= beta )
					{
					int bestValue = alpha;
						alpha = value; // test
						if ( alpha >= CHECKMATE_VALUE - MAX_SEARCH_DEPTH )
							alpha--; // test! //should be more efficient when detecting
						beta = -IMPOSSIBLE_VALUE;

//--- A Better Move has been found ---------------------------------------------//
						if ( m_bAnalyzeMode )
							cout << "++" << endl;

						if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
						{	
							value = core_alphabeta( stackL2, -beta, -alpha, 1, currentDepth, 
								   -(branchValue + attackvalue), zeroMove2 ); 
						}

						DEBUG_OUTPUT("\tValue failhigh 2: %i", value);
						DEBUG_OUTPUT("New move from %i to %i\n",m->sq1, m->sq2);

						if ( value > bestValue )
						{
							alpha = value;
							beta = alpha + 1;
							DEBUG_OUTPUT("Better move detected!");

							if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
							{
								extendPVAR(0, m);
								postThinking(value);
							}
						}
						else
						{
							alpha = bestValue;
							beta = alpha + 1;
						}

						if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
							m->value = value; 
						else // dont do an update, if the engine has been stopped
							m->value = -CHECKMATE_VALUE;
					}
					else
					{
						if ( (!m_bStopEngine && !softStop) || m_baseDepth == 2)
							m->value = value - 1; 
						else // dont do an update, if the engine has been stopped
							m->value = -CHECKMATE_VALUE;
					}
				}

				popMove();
				first = false;
			}
		}	
		stackL1.BestOrientation();
	}
}

int Engine::core_alphabeta( NodeStack &stackL1, int alpha, int beta, int depth, 
							int finalDepth, int branchValue, bool zeroMove)
{
	m_nodeCount++; 
	m_PV[depth][depth][0] = NO_SQUARE; // initialize principle variant


//--- Winbord asked for an update -----------------------------------------------//
	if (m_bPostAnalysis)
	{
		postAnalysis();
		m_bPostAnalysis = false;
	}
//--- Reasons to stop search ----------------------------------------------------//
	verifyStopConditions();
	if ( m_bStopEngine && m_baseDepth != 2) 
		return -CHECKMATE_VALUE;
	if ( m_drawValue < CRITICAL_DRAW_VALUE ) 
		return 0; // draw (insufficient material)
	if ( alpha >= CHECKMATE_VALUE - depth - 1 ) //test!
		return branchValue;
	if ( m_hash->GetConfigurationCount() >/*>=*/ 2 && m_zeroMoveCount == 0 ) //test        
		return 0;

bool checkmate = true, check = false;

	check = m_pKing[m_move.toPlay]->isAttacked();

int posValue = m_evaluator.GetPositionValue(alpha, beta, branchValue);

//--- Razoring ------------------------------------------------------------------//
	if ( !check && zeroMove && depth + 1 >= finalDepth && 
		 posValue - Q_PAWN > beta ) // One pawn security
		return - posValue; 

//--- Check extension -----------------------------------------------------------//
	if ( check && finalDepth < 2 * m_baseDepth )
		finalDepth++; // check, lets have a closer look 

ExtendedMove* last = m_gameHistory->GetTail();

//--- Promotion extension -------------------------------------------------------//
	if ( last->pP1->type == PAWN && finalDepth < 2 * m_baseDepth ) 
		if ( ( last->pP1->color == WHITE && (last->sq1 >> 3) > 3 ) ||  
			 ( last->pP1->color == BLACK && (last->sq1 >> 3) < 4 ))  
			finalDepth++;

//--- Negative promotion extension ----------------------------------------------//
	if ( last->pP2->type == PAWN )
		if ( ( last->pP2->color == WHITE && (last->sq2 >> 3) > 4) || 
			 ( last->pP2->color == BLACK && (last->sq2 >> 3) < 3) )
			finalDepth--;

//--- Consult hash table --------------------------------------------------------//
	switch(m_hash->GetType(finalDepth-depth, m_move.toPlay))
	{
	case NO_TYPE: break;
	case IN_WINDOW: return -m_hash->GetValue(); break;
	case FAIL_LOW:
		if ( m_hash->GetValue() < alpha ) return -m_hash->GetValue();
		break;
	case FAIL_HIGH:
		if ( m_hash->GetValue() >= beta ) return -m_hash->GetValue();
		break;
	}

//--- Generate Moves ------------------------------------------------------------//
bool zeroMove2 = generateAllMoves(stackL1, check, depth);
int value = 0, bestValue = IMPOSSIBLE_VALUE, attackvalue = 0, depthBackup;
Move* m;
Move* bestInWindow = NULL;
Move* bestFailLow = NULL;
NodeStack stackL2(stackL1.GetTail()); 

//--- Razoring with zero move information ---------------------------------------//
	if ( !check && zeroMove && zeroMove2 
		 && depth + 1 >= finalDepth && posValue > beta ) 
		return - posValue; 

//--- Nullmove ------------------------------------------------------------------//
	if ( zeroMove && zeroMove2 && posValue > beta &&
		 depth + 2 < finalDepth && 
		 beta >= - CHECKMATE_VALUE + MAX_SEARCH_DEPTH ) 
	{
		value = nullmove(&stackL2, alpha, beta, 
			             depth + 2, finalDepth, branchValue);

		if ( value >= beta ) 
		{
			m_move.sq1 = NO_SQUARE;
			if ( !m_bStopEngine )
			{
				m_hash->EnterValue(&m_move, value, //test, ev FAIL_HIGH_NULL
						   finalDepth - depth, FAIL_HIGH);
			}
			return - value; 
		}

//--- Threat extension ----------------------------------------------------------//
		if ( value < -CHECKMATE_VALUE + MAX_SEARCH_DEPTH &&
			 finalDepth < 2 * m_baseDepth ) 
			finalDepth++;
	}

//--- Core alpha beta -----------------------------------------------------------//
	depthBackup = finalDepth;

	while ((m = stackL1.GetNextBest()) != NULL 
		  && alpha != (CHECKMATE_VALUE - depth - 1) ) //TEST! might cause problems
	{						
		m_move = *m;

//--- Capture / Recapture extension ---------------------------------------------//
		finalDepth = depthBackup; // test, this extension is not yet proberly 
		if ( last->sq2 == m_move.sq2 &&	// treated in the hash entry
			 last->pP2->value + Q_PAWN > m_board[m_move.sq2]->value &&
			 last->pP2->value - Q_PAWN < m_board[m_move.sq2]->value && 
			 finalDepth < 2 * m_baseDepth ) 
			finalDepth++;

		attackvalue = pushMove();

		// Important: Special treatement for the en passant move
		if ( ( m->check_test || m->special == EN_PASSANT ) 
			 && ( !m->valid || m_pKing[!m_move.toPlay]->isAttacked() ) )
		{
			m->valid = false;
			m->value = IMPOSSIBLE_VALUE;
			popMove();
		}
		else
		{
			checkmate = false;
			m->check_test = false;

			if ( depth + 1 < finalDepth ) 
			{
				value = core_alphabeta( stackL2, -beta, -alpha, depth + 1, finalDepth, 
						       -(branchValue + attackvalue), zeroMove2 ); 
			}
			else
			{
				m_PV[depth + 1][depth + 1][0] = NO_SQUARE;
				value = extension_alphabeta( stackL2, -beta, -alpha, 0, finalDepth, 
					                  -(branchValue + attackvalue) ); 			
									  //-(posValue + attackvalue));
			}

			m->value = value;
			popMove();

			if( value > bestValue ) 
			{ 
				bestValue = value; 
				bestFailLow = m;

				if( value >= beta )
				{   
					extendPVAR(depth, m);

					m_killer[depth][1][0] = m_killer[depth][0][0];
					m_killer[depth][1][1] = m_killer[depth][0][1];
					m_killer[depth][0][0] = m->sq1;
					m_killer[depth][0][1] = m->sq2;

					m_board[m->sq1]->bonus[m->sq2] += 1;

					if ( !m_bStopEngine )
					{
						m_hash->EnterValue(m, value,
								finalDepth - depth, FAIL_HIGH);
					}

					return - value;
				}
				else if( value > alpha )
				{  
					extendPVAR(depth, m);
					bestInWindow = m;
					alpha = value;   
				}
			}
		}
	}	
	
//--- Detect Checkmate ---------------------------------------------------------//
	if ( checkmate )
	{
		if ( check ) return CHECKMATE_VALUE - depth;
		else return 0; 
	}
	
//--- Process statistics and hash table entries --------------------------------//
	if ( bestInWindow != NULL ) 
	{
		m_board[bestInWindow->sq1]->bonus[bestInWindow->sq2] += 4; // some kind of history
																   // heuristic
		if ( !m_bStopEngine )									
		{
			m_hash->EnterValue(bestInWindow, bestInWindow->value, 
						   finalDepth - depth, IN_WINDOW);
		}
	}
	else if ( bestFailLow != NULL )
	{
		if ( !m_bStopEngine )
		{
			m_hash->EnterValue(bestFailLow, bestFailLow->value,
						   finalDepth - depth, FAIL_LOW);
		}
	}
	
	return -bestValue;
}

int Engine::extension_alphabeta(NodeStack& stackL1, int alpha, int beta, int extensionDepth, 
						  int finalDepth, int branchValue )
{
	m_nodeCount++; 

	verifyStopConditions();
	if ( m_bStopEngine && m_baseDepth != 2) 
		return -CHECKMATE_VALUE;
	if ( m_drawValue < CRITICAL_DRAW_VALUE ) return 0;//-branchValue; 

int posValue = m_evaluator.GetPositionValue(alpha, beta, branchValue);

	if ( extensionDepth >= 6 /*4*/ || posValue >= beta )
		return -posValue;

int value = 0, bestValue = posValue, attackvalue = 0; 

	generateAllXMoves(stackL1);
NodeStack stackL2(stackL1.GetTail());
Move* m;

	while ( (m = stackL1.GetNextBest()) != NULL )
	{
		m_move = *m;
		
		attackvalue = pushMove();

		if ( !m_pKing[!m_move.toPlay]->isAttacked() )
		{			
			value = extension_alphabeta( stackL2, -beta, -alpha, extensionDepth + 1,
								  finalDepth, -(branchValue + attackvalue) );

			if( value >= bestValue ) 
			{ 
				bestValue = value; 
				if( value >= beta )
				{   
					popMove();
					return -value;
				}
				else if ( value > alpha )
				{  
					alpha = value;   
				}
			}
		}

		popMove();
	}	

	return -bestValue;
}

void Engine::extendPVAR(int depth, Move* best)
{
	m_PV[depth][depth][0] = best->sq1;
	m_PV[depth][depth][1] = best->sq2;

#ifndef NDEBUG
	if ( depth == 0 ) 
	{
		DEBUG_OUTPUT("\nBest line: %i->%i; ", m_PV[0][0][0], m_PV[0][0][1]);
	}
#endif

	for( int i = 1; i < MAX_SEARCH_DEPTH; i++ ) 
	{
		if( m_PV[depth + 1][depth + i][0] == NO_SQUARE ) 
		{
			m_PV[depth][depth + i][0] = NO_SQUARE;
			return;
		}

		m_PV[depth][depth + i][0] =
				m_PV[depth + 1][depth + i][0];
		m_PV[depth][depth + i][1] =
				m_PV[depth + 1][depth + i][1];

#ifndef NDEBUG
		if ( depth == 0 )
		{
			DEBUG_OUTPUT("; %i->%i ", m_PV[0][i][0], m_PV[0][i][1]);
		}
#endif
	}
}

void Engine::postThinking(int value)
{
	if ( !m_bPostThinking )
		return;

	string s = "";
	for( int i = 0; i < MAX_SEARCH_DEPTH; i++ ) 
	{
		s += " " + MoveToString(m_PV[0][i][0],m_PV[0][i][1], NOTHING);
		if ( m_PV[0][i][0] == NO_SQUARE)
		{
			long time = GetTickCount(); 
			cout << m_baseDepth << " " << (value / 10) << " " 
		         << (time - m_timeOffset)/10 
			     << " " << m_nodeCount << s << endl;
			return;
		}
	}
}

void Engine::postAnalysis()
{
	long time = GetTickCount();
	cout << "stat01: " << (time - m_timeOffset)/10 << " "
		 << m_nodeCount << " " << m_baseDepth << " " 
		 << (m_analysisMvTot - m_analysisMvDone) << " " << m_analysisMvTot << endl; 
}


int Engine::nullmove(NodeStack* stackL1, int alpha, int beta, int depth,
						int finalDepth, int branchValue)
{
int value;
int ep = m_epSquare;
	m_epSquare = NO_SQUARE;
	m_move.toPlay = !m_move.toPlay;
	m_zeroMoveCount++;

	//if ( depth >= finalDepth )
	//	value = extension_alphabeta( -beta, -alpha, 0, finalDepth, -branchValue ); 
	//else
	value = core_alphabeta( *stackL1, -beta, -alpha, depth, finalDepth, -branchValue, false); 

	m_zeroMoveCount--;
	m_move.toPlay = !m_move.toPlay;
	m_epSquare = ep;

	return value;
}

void Engine::InitializeEvaluation(int& branchValue)
{
Piece* p; 

	for ( int fNr = 0; fNr < 32; fNr++ )
	{
		for ( int square = 0; square < 64; square++ )
			m_pieceList[fNr]->bonus[square] = 0;
	}

	p = m_pKing[m_move.toPlay];
	do
	{
		branchValue += p->value;
		p = p->next;
	} while ( p->type != KING );

	p = m_pKing[!m_move.toPlay];
	do
	{
		branchValue -= p->value;
		p = p->next;
	} while ( p->type != KING );

	m_evaluator.SetPositionType(AUTO_CHOOSE_POSITION_TYPE, true);
	m_evaluator.SetEvaluationWindowOffset();

	m_drawValue = 0;
	for ( int square = 0; square < 64; square++ ) 
	{
		m_drawValue += m_board[square]->drawValue;
		m_pieceList[NO_PIECE_NR]->bonus[square] = 0;
	}

	for ( int i = 0; i < MAX_SEARCH_DEPTH; i++ )
	{
		m_killer[i][0][0] = NO_SQUARE;
		m_killer[i][1][0] = NO_SQUARE;
	}
}

