/***************************************************************************
                     engine.h  -  tree search for ChessterfieldCL
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GNU_VERSION
#include <windows.h>
#endif

#include <ctime>

class Engine:public EngineObject  
{
private:
	void verifyStopConditions()
	{		
		if ( m_bAnalyzeMode || m_searchMode != TIME_LIMIT ) return;
		long time = GetTickCount();
		if ( time > m_timeSoftStop ) m_bSoftStop = true;
		if ( time > m_timeHardStop ) m_bStopEngine = true;
	};
	void postThinking(int value);
	void postAnalysis();
	string MoveToString(int sq1, int sq2, int special);

	void InitializeEvaluation(int& branchValue);
	void extendPVAR(int depth, Move* best);

	void root_alphabeta(NodeStack &stackL1, int searchDepth);
	int core_alphabeta( NodeStack &stackL1, int alpha, int beta, int depth, 
				int finalDepth, int branchValue, bool zeroMove);
    int extension_alphabeta(NodeStack& stackL1, int alpha, int beta, int extensionDepth, 
				int finalDepth, int branchValue);
	int nullmove(NodeStack* stackL1, int alpha, int beta, int depth, 
				int finalDepth, int branchValue);

	bool generateAllMoves(NodeStack &stack, bool check, int depth);
	void generateAllXMoves(NodeStack &stack);


	Evaluation m_evaluator;

	// stop conditions for timed search
	long m_timeSoftStop; // stop if it "makes sense"
	long m_timeHardStop; // stop the engine as fast as possible
	
	// for some analysis output:
	long m_timeOffset;
	int m_analysisMvDone;
	int m_analysisMvTot;
	int m_nodeCount;

	// parameters for the search:
	int m_baseDepth;
	int m_zeroMoveCount;
	int m_PV[MAX_SEARCH_DEPTH][MAX_SEARCH_DEPTH][2]; //[][][0] is sq1 square
												     //[][][1] is sq2 square
	int m_killer[MAX_SEARCH_DEPTH][2][2]; //[search depth][killer nr][sq1 | sq2]

	// first move of the principle variant
	Move m_pv1Move;
public:
	Engine(int* board, bool toPlay);
	~Engine();

	bool MakeComputerMove(long tSoftStop, long tHardStop, 
						  int searchDepth, int searchMode, bool bResetHash);
	bool MakeUserMove(int from, int to, int castlingType);
	bool UndoMove();

	bool toPlay() {return m_move.toPlay;};
	void setToPlay(bool color) { m_move.toPlay = color;};
	string GetLastMove() { return MoveToString(m_move.sq1, m_move.sq2, m_move.special);};
	string GetPV1Move() 
	{ 
		if ( !m_pv1Move.valid ) return "";
		else return MoveToString(m_pv1Move.sq1, m_pv1Move.sq2, m_pv1Move.special);
	};
	int GetState() {return m_move.remark;};


	//Engine settings:
	bool m_bRequestHash;
	bool m_bPostThinking;
	bool m_bAnalyzeMode;
	bool m_bPostAnalysis;

	bool m_bStopEngine;
	bool m_bSoftStop;
	int m_searchMode;
};