/***************************************************************************
         engineobject.cpp  -  baseclass for most Chessterfield objects
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "def_inc.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "hashtable.h"

#include <iostream>
#include <cassert>

// EngineObject
//
/////////////////////////////////////////////////////////////////////////////

Move         EngineObject::m_move;
GameHistory* EngineObject::m_gameHistory;
HashTable*   EngineObject::m_hash;
Piece*       EngineObject::m_board[65];
Piece*       EngineObject::m_pieceList[33];
int          EngineObject::m_epSquare = NO_SQUARE;
NodeStack*   EngineObject::m_currentStack;
Move*		 EngineObject::m_moveStack;
bool         EngineObject::m_staticInitDone = false;
King*        EngineObject::m_pKing[2] = {NULL, NULL};

int  	  	 EngineObject::m_wPawnColumns[10];
int  		 EngineObject::m_bPawnColumns[10]; 
bool		 EngineObject::m_pawnEndgame;
int		     EngineObject::m_drawValue;
int			 EngineObject::m_positionType;
bool		 EngineObject::m_kingFields1[64];
bool		 EngineObject::m_kingFields2[64];
bool*	     EngineObject::m_owKingFields;
bool*	     EngineObject::m_opKingFields;


const int EngineObject::m_initPawnColumns[10] = {0,0,0,0,0,0,0,0,0,0};

EngineObject::EngineObject()
{
	if ( !m_staticInitDone )
	{
		m_staticInitDone = true;

		m_hash = new HashTable;
		m_gameHistory = new GameHistory;
#ifdef EN_VERSION
		m_moveStack = new Move[100];
#else
		m_moveStack = new Move[1200];
#endif

		m_positionType = OPENING; 
		m_pawnEndgame = false;

		m_currentStack = NULL;
	}
}

void EngineObject::DeallocateStaticMembers()
{
	m_staticInitDone = false;
	DeleteAllPieces();
	delete m_hash;
	delete m_gameHistory;
	delete [] m_moveStack;
}

void EngineObject::DeleteAllPieces()
{
	// to make sure that the castling pieces are deleted:
	while ( m_gameHistory->GetCount() > 0 ) popMove(); 

	for ( int i = 0; i < 33; i++ ) 
	{
		if ( m_pieceList[i] != NULL ) 
		{
			delete m_pieceList[i];
			m_pieceList[i] = NULL;
		}
	}
}

int EngineObject::pushMove()
{
int attackvalue;
ExtendedMove *pMove;
	m_hash->PushIndexKey();

	if ( m_board[m_move.sq2]->type != NO_PIECE ) 
	{
		m_hash->AddRemoveFigure(m_move.sq2);
		attackvalue = m_board[m_move.sq2]->value;
		m_drawValue -= m_board[m_move.sq2]->drawValue;
		m_board[m_move.sq2]->previous->next = m_board[m_move.sq2]->next;
		m_board[m_move.sq2]->next->previous = m_board[m_move.sq2]->previous;
	}
	else attackvalue = 0;

	if ( m_board[m_move.sq2]->type == KING )
		assert(false);

	pMove = m_gameHistory->AddElement();

	m_board[m_move.sq2]->moveCount++;
	m_board[m_move.sq1]->moveCount++;
	pMove->sq1 = m_move.sq1;
	pMove->pP1 = m_board[m_move.sq1];
	pMove->sq2 = m_move.sq2;
	pMove->pP2 = m_board[m_move.sq2];
	pMove->epSquare = m_epSquare;
	m_epSquare = NO_SQUARE;
	m_board[m_move.sq1]->square = m_move.sq2;
	m_board[m_move.sq2] = m_board[m_move.sq1];
	m_board[m_move.sq1] = m_pieceList[32];

	m_hash->MoveFigure(m_move.sq1, m_move.sq2);

	switch ( m_move.special )
	{
		case NOTHING:
			pMove->special = NOTHING;
		break;
		case DOUBLE_ADVANCE:
			m_epSquare = (m_move.sq1 + m_move.sq2)>>1;
			pMove->special = NOTHING;
		break;
		case EN_PASSANT:
			attackvalue = Q_PAWN;
			m_drawValue -= DRAW_VALUE_PAWN;
			pMove->special = EN_PASSANT;
			if ( m_board[m_move.sq2]->color == WHITE )
			{
				m_hash->AddRemoveFigure(m_move.sq2 - 8);
				pMove->sq3 = m_move.sq2 - 8;
				pMove->pP3 = m_board[m_move.sq2 - 8];
				m_board[m_move.sq2 - 8]->previous->next = m_board[m_move.sq2 - 8]->next;
				m_board[m_move.sq2 - 8]->next->previous = m_board[m_move.sq2 - 8]->previous;
				m_board[m_move.sq2 - 8] = m_pieceList[32];
			}
			else
			{
				m_hash->AddRemoveFigure(m_move.sq2 + 8);
				pMove->sq3 = m_move.sq2 + 8;
				pMove->pP3 = m_board[m_move.sq2 + 8];
				m_board[m_move.sq2 + 8]->previous->next = m_board[m_move.sq2 + 8]->next;
				m_board[m_move.sq2 + 8]->next->previous = m_board[m_move.sq2 + 8]->previous;
				m_board[m_move.sq2 + 8] = m_pieceList[32];
			}
		break;
		case QUEEN_CASTLING:
			m_pKing[m_move.toPlay]->m_bCastled = true;
			//attackvalue += 50;
			pMove->special = QUEEN_CASTLING;
			pMove->sq3 = m_move.sq1 - 4;
			pMove->pP3 = m_board[m_move.sq1 - 4];
			pMove->sq4 = m_move.sq1 - 1;
			m_board[m_move.sq1 - 1] = m_board[m_move.sq1 - 4];
			m_board[m_move.sq1 - 1]->square = m_move.sq1 - 1;
			m_board[m_move.sq1 - 4] = m_pieceList[32];
			m_board[m_move.sq1 - 1]->moveCount++;
			m_hash->MoveFigure(m_move.sq1 - 4, m_move.sq1 - 1);
		break;
		case KING_CASTLING:
			m_pKing[m_move.toPlay]->m_bCastled = true;
			//attackvalue += 50;
			pMove->special = KING_CASTLING;
			pMove->sq3 = m_move.sq1 + 3;
			pMove->pP3 = m_board[m_move.sq1 + 3];
			pMove->sq4 = m_move.sq1 + 1;
			m_board[m_move.sq1 + 1] = m_board[m_move.sq1 + 3];
			m_board[m_move.sq1 + 1]->square = m_move.sq1 + 1;
			m_board[m_move.sq1 + 3] = m_pieceList[32];
			m_board[m_move.sq1 + 1]->moveCount++;
			m_hash->MoveFigure(m_move.sq1 + 3, m_move.sq1 + 1);
		break;
		case NEW_QUEEN:
			attackvalue += Q_QUEEN - Q_PAWN;
			pMove->special = NEW_QUEEN;
			m_hash->AddRemoveFigure(m_move.sq2);
			pMove->pP3 = m_board[m_move.sq2];
			pMove->pP1 = new Queen(0, 0, 0, m_board[m_move.sq2]);
			m_hash->AddRemoveFigure(m_move.sq2);
		break;
		case NEW_KNIGHT:
			attackvalue += Q_KNIGHT - Q_PAWN;
			m_drawValue += DRAW_VALUE_KNIGHT - DRAW_VALUE_PAWN;
			pMove->special = NEW_KNIGHT;
			m_hash->AddRemoveFigure(m_move.sq2);
			pMove->pP3 = m_board[m_move.sq2];
			pMove->pP1 = new Knight(0, 0, 0, m_board[m_move.sq2]);
			m_hash->AddRemoveFigure(m_move.sq2);
		break;
		case NEW_BISHOP:
			attackvalue += Q_BISHOP - Q_PAWN;
			m_drawValue += DRAW_VALUE_BISHOP - DRAW_VALUE_PAWN;
			pMove->special = NEW_BISHOP;
			m_hash->AddRemoveFigure(m_move.sq2);
			pMove->pP3 = m_board[m_move.sq2];
			pMove->pP1 = new Bishop(0, 0, 0, m_board[m_move.sq2]);
			m_hash->AddRemoveFigure(m_move.sq2);
		break;
		case NEW_ROOK:
			attackvalue += Q_ROOK - Q_PAWN;
			pMove->special = NEW_ROOK;
			m_hash->AddRemoveFigure(m_move.sq2);
			pMove->pP3 = m_board[m_move.sq2];
			pMove->pP1 = new Rook(0, 0, 0, m_board[m_move.sq2]);
			m_hash->AddRemoveFigure(m_move.sq2);
		break;
	}

	m_move.toPlay = !m_move.toPlay;
	m_hash->IncrementConfigurationCount();

	return attackvalue;
}

void EngineObject::popMove()
{
ExtendedMove* pMove;
	m_hash->DecrementConfigurationCount();
	m_hash->PopIndexKey();
	pMove = m_gameHistory->GetTail();
	Piece *pFig1 = pMove->pP1, *pFig2 = pMove->pP2;

	m_board[pMove->sq1] = pFig1;
	m_board[pMove->sq2] = pFig2;

	if ( pFig2->type != NO_PIECE )
	{
		pFig2->previous->next = pFig2;
		pFig2->next->previous = pFig2;
	}

	pFig1->square = pMove->sq1;
	pFig2->square = pMove->sq2;
	pFig1->moveCount--;
	pFig2->moveCount--;
	m_epSquare = pMove->epSquare;
	m_drawValue += pFig2->drawValue;

	switch( pMove->special )
	{
		case NOTHING: break;
		case QUEEN_CASTLING:
			m_pKing[!m_move.toPlay]->m_bCastled = false;
			m_board[pMove->sq3] = m_board[pMove->sq4];
			m_board[pMove->sq4] = m_pieceList[32];
			m_board[pMove->sq3]->square = pMove->sq3;
			pMove->pP3->moveCount--; break;
		case KING_CASTLING:
			m_pKing[!m_move.toPlay]->m_bCastled = false;
			m_board[pMove->sq3] = m_board[pMove->sq4];
			m_board[pMove->sq4] = m_pieceList[32];
			m_board[pMove->sq3]->square = pMove->sq3;
			pMove->pP3->moveCount--; break;
		case NEW_QUEEN:
			undoCastling();
			break;
		case NEW_KNIGHT: 
			m_drawValue += DRAW_VALUE_PAWN - DRAW_VALUE_KNIGHT;
			undoCastling();
			break;
		case NEW_BISHOP: 
			m_drawValue += DRAW_VALUE_PAWN - DRAW_VALUE_BISHOP;
			undoCastling();
			break;
		case NEW_ROOK: 
			undoCastling();
			break;
		case EN_PASSANT:
			m_drawValue += DRAW_VALUE_PAWN; 
			m_board[pMove->sq3] = pMove->pP3;
			pMove->pP3->square = pMove->sq3;
			pMove->pP3->previous->next = pMove->pP3;
			pMove->pP3->next->previous = pMove->pP3;
			break;
	}

	m_move.toPlay = !m_move.toPlay;
	m_gameHistory->RemoveElement();
}

bool EngineObject::IsValidMove()
{
	m_move.valid = true;

	if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
	{
		m_move.remark = CHECKMATE;
		m_move.valid = false;
	}

	if ( m_move.sq1 == m_move.sq2 || 
		 m_move.sq1 < 0 || m_move.sq1 > 63 ||
		 m_move.sq2 < 0 || m_move.sq2 > 63 || 
		 (m_board[m_move.sq1]->color == WHITE && m_move.toPlay == BLACK) ||
		 (m_board[m_move.sq1]->color == BLACK && m_move.toPlay == WHITE) ||
		 m_board[m_move.sq1]->type == NO_PIECE )
	{
		m_move.remark = NEW_MOVE;
		m_move.valid = false;
	}

	if ( m_move.valid )
	{
		int special = m_move.special;
		m_board[m_move.sq1]->verifyMove(); // Attention: this function overrides 
										   // m_move.special
		if ( m_move.special == NEW_QUEEN && ( special == NEW_ROOK || 
			 special == NEW_BISHOP || special == NEW_KNIGHT ))
			m_move.special = special;
	}

	if ( m_move.valid )
	{
		pushMove();
		if( m_pKing[m_board[m_move.sq2]->color]->isAttacked() )
		{
			m_move.valid = false;
			m_move.remark = NEW_MOVE;
		}
		else
			m_move.remark = DO_MOVE;
		popMove();
	}
	else
		m_move.remark = NEW_MOVE;

	return m_move.valid;
}

void EngineObject::undoCastling()
{
ExtendedMove* pMove = m_gameHistory->GetTail();
Piece* oldP = pMove->pP1;
Piece* newP = pMove->pP3;
	m_pieceList[oldP->number] = newP;
	m_board[oldP->square] = newP;
	newP->square = oldP->square;
	newP->moveCount--;
	newP->next = oldP->next;
	newP->previous = oldP->previous;
	newP->next->previous = newP;
	newP->previous->next = newP;
	memcpy(newP->bonus, oldP->bonus, 64 * sizeof(int));
	delete oldP;
}

bool EngineObject::DoPGNMove(int from_r, int from_c, int to_r, 
						     int to_c, int piece, int p_piece )
{
	if ( m_gameHistory->GetCount() + 2 >= MAX_NR_OF_MOVES )
	{
		m_move.remark = MOVECOUNT_LIMIT;
		m_move.valid = false;
		return false;
	}

	if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
	{
		m_move.remark = CHECKMATE;
		m_move.valid = false;
		return false;
	}

	NodeStack nstack(m_moveStack);
	m_currentStack = &nstack;

	Piece* p = m_pKing[m_move.toPlay];

	do
	{
		if ( p->type == piece ) 
		{
			p->generateMoves();
		}
		p = p->next;
	} while ( p->type != KING );

	Move* m = m_moveStack;
	Move* t = nstack.GetTail();
	bool moveFound = false;

	while ( m < t && !moveFound )
	{
		moveFound = true;
		if ( from_r != 8 && (m->sq1 >> 3) != from_r ) moveFound = false;
		if ( from_c != 8 && (m->sq1 & 7) != from_c ) moveFound = false;
		if ( to_r != 8 && (m->sq2 >> 3) != to_r ) moveFound = false;
		if ( to_c != 8 && (m->sq2 & 7) != to_c ) moveFound = false;
		if ( !moveFound ) m++;
	}

	if ( !moveFound ) 
		return false;

	switch ( p_piece )
	{
	case QUEEN: m->special = NEW_QUEEN; break;
	case ROOK: m->special = NEW_ROOK; break;
	case BISHOP: m->special = NEW_BISHOP; break;
	case KNIGHT: m->special = NEW_KNIGHT; break;
	}

	m_move = *m;

	pushMove();

	if ( m_pKing[!m_move.toPlay]->isAttacked() ) 
	{
		popMove();
		m_move.valid = false;
		return false;
	}
	else
		return true;
}
