/***************************************************************************
          evaluation.h  -  evaluates positions with network.dat file
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <vector>

using namespace std;

#ifdef EN_VERSION
#define WEIGHTTYPE float 
#else 
#define WEIGHTTYPE int
#endif

class Configuration
{
public:
	unsigned int bitcode;
	unsigned int not_msac; // not most specific active configuration
	WEIGHTTYPE weight;
	int jump;
	int IsActive(unsigned int pbc){ return ((bitcode & pbc) == bitcode); };
	int Jump(unsigned int pbc)
	{ 
		if ( not_msac & pbc ) return 1;
		else return jump;
	};
#ifdef EN_VERSION
	int count; //For preconditioning
	double ngrad0; //-gradient(k)
	double ngrad1; //-gradient(k+1)
	double preconditioner; //Jacobi preconditioning
#endif //EN_VERSION
};


typedef vector<Configuration> ConfigurationSet;
typedef ConfigurationSet::iterator ConfigurationSetIt;

class TableEntry
{
public:
	TableEntry() { Clear(); };
	void Clear() { bitcode = 0; value = 0; };
	unsigned int bitcode;
	WEIGHTTYPE value;
};

#ifdef BIG_BITCODE_TABLE
#define TABLE_SIZE 65536
#define TABLE_MASK 65535
#else
#define TABLE_SIZE 256
#define TABLE_MASK 255
#endif

class BitcodeTable
{
private:
	TableEntry* table[6];
public:
	BitcodeTable()
	{
		for ( int i = 0; i < 6; i++ )
			table[i] = new TableEntry[TABLE_SIZE];
	};
	~BitcodeTable()
	{
		for ( int i = 0; i < 6; i++ )
			delete [] table[i];
	};
	bool GetEntry(unsigned int bitcode, int piece, WEIGHTTYPE& value)
	{
		unsigned int index;
#ifdef BIG_BITCODE_TABLE
		index = (bitcode>>16&TABLE_MASK)^(bitcode&TABLE_MASK);
#else
		index = (bitcode>>24&TABLE_MASK)^(bitcode>>16&TABLE_MASK)
		        ^(bitcode>>8&TABLE_MASK)^(bitcode&TABLE_MASK);
#endif

		TableEntry& te = table[piece][index]; 
		if ( te.bitcode == bitcode )
		{
			value += te.value;
			return true;
		}
		else return false;
	};
	void AddEntry(unsigned int bitcode, int piece, WEIGHTTYPE value)
	{
		unsigned int index;
#ifdef BIG_BITCODE_TABLE
		index = (bitcode>>16&TABLE_MASK)^(bitcode&TABLE_MASK);
#else
		index = (bitcode>>24&TABLE_MASK)^(bitcode>>16&TABLE_MASK)
		        ^(bitcode>>8&TABLE_MASK)^(bitcode&TABLE_MASK);
#endif
		TableEntry& te = table[piece][index]; 
		te.bitcode = bitcode;
		te.value = value;
	};
	void Clear()
	{
		for ( int i = 0; i < 6; i++ )
			for ( int j = 0; j < TABLE_SIZE; j++ ) table[i][j].Clear();
	};
};

class Evaluation:public EngineObject
{
public:
	Evaluation();
	void SetEvaluationWindowOffset();
	int GetPositionValue(int alpha, int beta, int branchValue);
	void GenerateAtomics();
	int SetPositionType(int posType, bool clear);
#ifdef EN_VERSION
	void SaveData();
	double Gstep(PositionSet& set);
	double CGstep(PositionSet& set, bool output);
	void InitWeights();
	float GetExampleValue(Position& pos);
	double GetAverageError(PositionSet& set);
	float GetHitPercentage(PositionSet& set);
	void RemoveConfigurations(int piece);
	int AddConfiguration(int piece);
	Configuration& GetConfiguration(int piece, int nr);
	void OptimizeConfigurationSet();
	int RemoveUnimportantConfigurations(ConfigurationSetIt& it, ConfigurationSet& set);
#endif // EN_VERSION

private:
	void LoadData();
	BitcodeTable tpBitcodeTable;
	BitcodeTable ieBitcodeTable;
	static ConfigurationSet configs[MAX_POSITIONTYPES][6]; 
	static bool staticInitDone;
	int evalWindowOffset[2];
	ConfigurationSet* currentConfigs;
#ifdef EN_VERSION
	float BuildPreconditioner(Position& pos);
	float GetExampleValue(Position& pos, double& posNGrad);
	void UpdateGradient(Position& pos, double error);
	double lastGradSquare;
#endif // EN_VERSION
};

