/***************************************************************************
        hashtable.cpp  -  hash table (also used for repetition rule)
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "def_inc.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "hashtable.h"

#include <iostream>

// shared memory
bool HashTable::m_bMemoryAvailable = true;
char* HashTable::m_distance = NULL;
char* HashTable::m_clearDistance = NULL;
char* HashTable::m_square1 = NULL;
char* HashTable::m_square2 = NULL;
unsigned char* HashTable::m_flags = NULL;
unsigned int* HashTable::m_key = NULL;
int* HashTable::m_value = NULL;

HashTable::HashTable()
{
int i, j;
	m_size = 524288L; //2 power 19
	m_andIndex = 524287L;
	m_bHasMemory = false;
	m_curKey = m_curKeyA;
	m_curIndex = m_curIndexA;
	m_curEntry = m_curEntryA;

	srand(1);//always the same random numbers

	for ( i = 0; i < 64; i++ )
	{
		for ( int j = 0; j < 12; j++ )
		{
			m_indexValue[i][j] = (rand() | (rand() << 15) | (rand() << 30));
			m_keyValue[i][j] = (rand() | (rand() << 15) | (rand() << 30));
		}
		m_indexValue[i][12] = 0;
		m_keyValue[i][12] = 0;
	}

	for ( i = 0; i < 512; i++ )
		for ( j = 0; j < 4; j++ )
		{
			m_repetitionKey[i][j] = 0;
			m_repetitionCount[i][j] = 0;
		}
}

HashTable::~HashTable()
{
	if ( m_bHasMemory || m_bMemoryAvailable )
	{

		m_bMemoryAvailable = false;
		AllocateMemory(false);
	}
	m_bMemoryAvailable = true;
}

bool HashTable::AllocateMemory(bool allocate)
{
	if ( allocate && m_distance == NULL )
	{
		if ( (m_distance = new char[m_size]) == NULL ) allocate = false;
		if ( (m_clearDistance = new char[m_size]) == NULL ) 
		{
			allocate = false;
		}
		else
		{
			for ( int i = 0; i < m_size; i++ )
				m_clearDistance[i] = -MAX_SEARCH_DEPTH;
		}
		if ( (m_square1 = new char[m_size]) == NULL ) allocate = false;
		if ( (m_square2 = new char[m_size]) == NULL ) allocate = false;
		if ( (m_flags = new unsigned char[m_size]) == NULL ) allocate = false;
		if ( (m_key = new unsigned int[m_size]) == NULL ) allocate = false;
		if ( (m_value = new int[m_size]) == NULL ) allocate = false;
	}

	if ( !allocate )
	{
		if ( m_square1 != NULL ) delete [] m_square1;
		if ( m_square2 != NULL ) delete [] m_square2;
		if ( m_flags != NULL ) delete [] m_flags;
		if ( m_key != NULL ) delete [] m_key;
		if ( m_value != NULL ) delete [] m_value;
		if ( m_distance != NULL ) delete [] m_distance;
		if ( m_clearDistance != NULL ) delete [] m_clearDistance;

		m_distance = NULL;
		m_clearDistance = NULL;
		m_square1 = NULL;
		m_square2 = NULL;
		m_flags = NULL;
		m_key = NULL;
		m_value = NULL;
	}

	return allocate;
}

void HashTable::CalculateKeyNIndex()
{
	(*m_curKey) = 0;
	(*m_curIndex) = 0;
	(*m_curEntry) = 0;

	for ( int i = 0; i < 64; i++ )
	{
		(*m_curIndex) = (*m_curIndex) ^ m_indexValue[i][m_board[i]->hashtype];
		(*m_curKey) = (*m_curKey) ^ m_keyValue[i][m_board[i]->hashtype];
	}
	(*m_curEntry) = ((*m_curIndex) & m_andIndex);
}

void HashTable::RequestHash(bool request, bool reset)
{
	if( request && m_bMemoryAvailable )
	{
		m_bMemoryAvailable = false;
		bool success = true;
		if ( m_distance == NULL )
		{
			success = AllocateMemory(true);

			if ( success ) // reset the hash table
			{
				for ( int i = 0; i < m_size; i++ )
				{
					m_square1[i] = NO_SQUARE;
					m_flags[i] = NO_TYPE;
					m_key[i] = 0;
					m_value[i] = IMPOSSIBLE_VALUE;
					m_distance[i] = -MAX_SEARCH_DEPTH;
				}
			}
		}

		if ( success && reset )
		{
			memcpy(m_distance, m_clearDistance, m_size * sizeof(char));
		}

		m_bHasMemory = success;
	}
	else
		m_bHasMemory = false;
}

void HashTable::ReleaseHash()
{
	if ( !m_bHasMemory ) return;
	m_bMemoryAvailable = true;
	m_bHasMemory = false;

//int count = 0;
	//for ( int i = 0; i < m_size; i++ )
	//{
	//	if ( m_value[i] != IMPOSSIBLE_VALUE )//test
	//		count++;//test
	//	m_value[i] = IMPOSSIBLE_VALUE;
	//}
	//DEBUG_OUTPUT("\nNumber of entries: %i/%i\n", count, m_nEntries);
}


