/***************************************************************************
              hashtable.h  -  hash table (also used for repetition rule)
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <iostream>

class Piece;

class HashTable:public EngineObject
{
public:
	HashTable();
	~HashTable();
	void RequestHash(bool request, bool reset);
	void ReleaseHash();

	void Initialize() { CalculateKeyNIndex(); };

	void IncrementConfigurationCount()
	{
	int ix = (*m_curIndex) & ((unsigned int) 511);
		if ( m_repetitionCount[ix][0] == 0 || m_repetitionKey[ix][0] == *m_curKey )
		{ 
			m_repetitionCount[ix][0]++; 
			m_repetitionKey[ix][0] = *m_curKey; 
		}
		else if ( m_repetitionCount[ix][1] == 0 || m_repetitionKey[ix][1] == *m_curKey )
		{ 
			m_repetitionCount[ix][1]++; 
			m_repetitionKey[ix][1] = *m_curKey; 
		}
		else if ( m_repetitionCount[ix][2] == 0 || m_repetitionKey[ix][2] == *m_curKey )
		{ 
			m_repetitionCount[ix][2]++; 
			m_repetitionKey[ix][2] = *m_curKey; 
		}
		else if ( m_repetitionCount[ix][3] == 0 || m_repetitionKey[ix][3] == *m_curKey )
		{ 
			m_repetitionCount[ix][3]++; 
			m_repetitionKey[ix][3] = *m_curKey; 
		}
	};

	void DecrementConfigurationCount()
	{
	int ix = (*m_curIndex) & ((unsigned int) 511);
		if ( m_repetitionKey[ix][0] == *m_curKey )
			m_repetitionCount[ix][0]--; 
		else if ( m_repetitionKey[ix][1] == *m_curKey )
			m_repetitionCount[ix][1]--; 
		else if ( m_repetitionKey[ix][2] == *m_curKey )
			m_repetitionCount[ix][2]--; 
		else if ( m_repetitionKey[ix][3] == *m_curKey )
			m_repetitionCount[ix][3]--; 
	};

	char GetConfigurationCount() const
	{
	int ix = (*m_curIndex) & ((unsigned int) 511);
		if ( m_repetitionCount[ix][0] == 0 || m_repetitionKey[ix][0] == *m_curKey )
			return m_repetitionCount[ix][0]; 
		else if ( m_repetitionCount[ix][1] == 0 || m_repetitionKey[ix][1] == *m_curKey )
			return m_repetitionCount[ix][1]; 
		else if ( m_repetitionCount[ix][2] == 0 || m_repetitionKey[ix][2] == *m_curKey )
			return m_repetitionCount[ix][2]; 
		else if ( m_repetitionCount[ix][3] == 0 || m_repetitionKey[ix][3] == *m_curKey )
			return m_repetitionCount[ix][3]; 
		else 
			return 0;
	};

	void EnterValue(Move* pZug, int value, char distance, char type )
	{
		if ( !m_bHasMemory || distance <= m_distance[*m_curEntry] )
			return;

//		if ( m_key[*m_curEntry] != (*m_curKey) )
//			m_nEntries++;//test only

/*#ifndef NDEBUG //test
	unsigned int key = (*m_curKey);
	unsigned int index = (*m_curIndex);
		CalculateKeyNIndex();
		if ( key != (*m_curKey) || index != (*m_curIndex) )
		{
			DEBUG_OUTPUT("Index or Key calculation failed!\n");
		}
#endif*/

		m_flags[*m_curEntry] = type;
		if ( pZug->toPlay )
			m_flags[*m_curEntry] += SIDE_TO_MOVE;
		m_distance[*m_curEntry] = distance;
		m_square1[*m_curEntry] = pZug->sq1;
		m_square2[*m_curEntry] = pZug->sq2;
		m_key[*m_curEntry] = (*m_curKey);
		m_value[*m_curEntry] = value;
	};
	int GetValue() //No access checking!!
	{
		return m_value[*m_curEntry];
	};
	char GetType(char distance, char sideToMove) const
	{	
		if ( !m_bHasMemory ||
			 m_key[*m_curEntry] != (*m_curKey) ||
			 m_distance[*m_curEntry] < distance || 
			 (m_flags[*m_curEntry] & SIDE_TO_MOVE) != sideToMove ||
			 m_value[*m_curEntry] == IMPOSSIBLE_VALUE )
			return NO_TYPE;
		else
			return (m_flags[*m_curEntry] & TYPE_MASK);
	};
	void GetMove(int& sq1, int& sq2) const
	{
		if ( !m_bHasMemory ||
			 m_key[*m_curEntry] != (*m_curKey))
		{
			sq1 = NO_SQUARE;
		}
		else
		{
/*#ifndef NDEBUG //test
		unsigned int key = (*m_curKey);
		unsigned int index = (*m_curIndex);
			CalculateKeyNIndex();
			if ( key != (*m_curKey) || index != (*m_curIndex) )
			{
				DEBUG_OUTPUT("Index or Key calculation failed!\n");
			}
#endif*/

			sq1 = m_square1[*m_curEntry];
			sq2 = m_square2[*m_curEntry];
		}
	};
	void AddRemoveFigure(int square)
	{
	int piece = m_board[square]->hashtype;
		(*m_curKey) ^= m_keyValue[square][piece];
		(*m_curIndex) ^= m_indexValue[square][piece];
		(*m_curEntry) = ((*m_curIndex) & m_andIndex);
	};

	void MoveFigure(int from, int to)
	{
	int piece = m_board[to]->hashtype;
		(*m_curKey) ^= m_keyValue[from][piece];
		(*m_curKey) ^= m_keyValue[to][piece];
		(*m_curIndex) ^= m_indexValue[from][piece];
		(*m_curIndex) ^= m_indexValue[to][piece];
		(*m_curEntry) = ((*m_curIndex) & m_andIndex);
	};
	void PopIndexKey()
	{
		m_curKey--;
		m_curIndex--;
		m_curEntry--;
	};
	void PushIndexKey()
	{
		m_curKey[1] = m_curKey[0];
		m_curIndex[1] = m_curIndex[0];
		m_curEntry[1] = m_curEntry[0];
		m_curKey++;
		m_curIndex++;
		m_curEntry++;
	};
private:
	// int m_nEntries;
	int m_size;
	unsigned int m_andIndex;
	bool m_bHasMemory;

	unsigned int* m_curKey;
	unsigned int* m_curIndex;
	unsigned int* m_curEntry;
	unsigned int m_curIndexA[MAX_NR_OF_MOVES + MAX_SEARCH_DEPTH + 1];
	unsigned int m_curEntryA[MAX_NR_OF_MOVES + MAX_SEARCH_DEPTH + 1];
	unsigned int m_curKeyA[MAX_NR_OF_MOVES + MAX_SEARCH_DEPTH + 1];
	unsigned int m_keyValue[64][13];
	unsigned int m_indexValue[64][13];

	unsigned int m_repetitionKey[512][4];
	char m_repetitionCount[512][4];
	// Shared Memory
	static bool m_bMemoryAvailable;
	static char* m_distance;
	static char* m_clearDistance;
	static char* m_square1;
	static char* m_square2;
	static unsigned char* m_flags;
	static unsigned int* m_key;
	static int* m_value;

	bool AllocateMemory(bool allocate);
	void CalculateKeyNIndex();
};

