/***************************************************************************
          learnobject.cpp  -  does the learning of games and endgames
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


// disable warning C4786: symbol greater than 255 character,
#ifndef GNU_VERSION
#pragma warning(disable: 4786) // Visual C++ stuff
#endif

#include <ctime> // To be removed

#include "def_inc.h"
// #include "mytcp.h"
#include "myvector.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "evaluation.h"
#include "pgnfile.h"

#include <string>
#include <fstream>
#include <iostream>
#include <vector>
#include <math.h>

using namespace std;

#include "learnobject.h"

LearnObject::LearnObject()
{
	for ( int i = 0; i < 33; i++ ) m_pieceList[i] = NULL;
	m_currentSet = NULL;
}

LearnObject::~LearnObject()
{
	DeallocateStaticMembers();
}

// filter options:
// ---------------

// 0	no filter is applied - all positions will be added to the training set

// 1	remove all positions which will be draw

// 2	positions with two bishops of the same color standing on squares with the
//		same color are removed 
//		Remark: useful filter for KBBK-endgame

// 3	positions that will soon lead to a checkmate or a promotion are removed

// 4	combination of filter 1 and filter 3

// 5	positions with exactly one bishop on a white square or
//      positions that will soon lead to a checkmate or a promotion are removed
//		Remark 1: this filter was written for the KBNK-endgame
//		Remark 2: this filter is not usful in the current implementation

bool LearnObject::FilterPosition(INTVECTOR* pieces, int value, 
							     bool toPlay, int filter)
{
	switch ( filter ) // see filter options
	{
	case 0: break;
	case 1: if ( value == 255 ) return false; break;
	case 2:	
		if (( pieces[W_BISHOP].size() == 2 &&
		   (((pieces[W_BISHOP][0] >> 3) + 
	   	     (pieces[W_BISHOP][0] & 7)) & 1 ) ==
		   (((pieces[W_BISHOP][1] >> 3) + 
			 (pieces[W_BISHOP][1] & 7)) & 1))
			  ||
			( pieces[B_BISHOP].size() == 2 &&
	  	   (((pieces[B_BISHOP][0] >> 3) + 
		     (pieces[B_BISHOP][0] & 7)) & 1 ) ==
		   (((pieces[B_BISHOP][1] >> 3) + 
			 (pieces[B_BISHOP][1] & 7)) & 1)))
			return false;
		break;
	case 3: if ( value <= 2 ) return false; break;
	case 4: if ( value <= 2 || value == 255 ) return false; break;
	case 5: 
		if (( pieces[W_BISHOP].size() == 1 &&
		   (((pieces[W_BISHOP][0] >> 3) + 
	   	     (pieces[W_BISHOP][0] & 7)) & 1 ) == 1)
			  ||
			( pieces[B_BISHOP].size() == 1 &&
	  	   (((pieces[B_BISHOP][0] >> 3) + 
		     (pieces[B_BISHOP][0] & 7)) & 1 ) == 1)
			 || value <= 6 )
			return false;
		break;
	}
	return true;
}

void LearnObject::AddCurrentPositionToSet(int value, bool toPlay)
{
	m_evaluator.GenerateAtomics();

PieceProperties ep;
Position pos;
Piece* p = m_pKing[toPlay];
	do
	{
		ep.atomics = p->atomics;
		ep.type = p->type;
		pos.tpPieces.push_back(ep);
		p = p->next;
	} while ( p->type != KING );

	p = m_pKing[!toPlay];
	do
	{
		ep.atomics = p->atomics;
		ep.type = p->type;
		pos.iePieces.push_back(ep);
		p = p->next;
	} while ( p->type != KING );

	pos.value = value;
	m_currentSet->push_back(pos);
}

void LearnObject::LoadPosition(INTVECTOR *pieces, bool toPlay)
{
	DeleteAllPieces();

	m_move.toPlay = toPlay;
	m_epSquare = NO_SQUARE;

int pNr, type;
IntVectorIt start, end, it;
int ptype[] = { PAWN, ROOK, BISHOP, KNIGHT, QUEEN, KING,
		    	PAWN, ROOK, BISHOP, KNIGHT, QUEEN, KING };

	new Piece(NO_COLOUR, NO_SQUARE, NO_PIECE_NR, NULL );
	for ( int square = 0; square < 65; square++ ) m_board[square] = m_pieceList[32];

	pNr = 15;
	for ( type = 11; type > 5; type-- )
	{
		start = pieces[type].begin(); end = pieces[type].end();

		for ( it = start; it != end; it++ )
		{
			switch ( ptype[type] )
			{
			case PAWN: new WPawn(WHITE, *it, pNr, NULL); break;
			case KNIGHT: new Knight(WHITE, *it, pNr, NULL); break;
			case BISHOP: new Bishop(WHITE, *it, pNr, NULL); break;
			case ROOK: new Rook(WHITE, *it, pNr, NULL); break;
			case QUEEN: new Queen(WHITE, *it, pNr, NULL); break;
			case KING: m_pKing[WHITE] = new King(WHITE, *it, pNr, NULL); break;
			}
			pNr--;
		}
	}

	pNr = 31;
	for ( type = 5; type >= 0; type-- )
	{
		start = pieces[type].begin(); end = pieces[type].end();

		for ( it = start; it != end; it++ )
		{
			switch ( ptype[type] )
			{
			case PAWN: new BPawn(BLACK, *it, pNr, NULL); break;
			case KNIGHT: new Knight(BLACK, *it, pNr, NULL); break;
			case BISHOP: new Bishop(BLACK, *it, pNr, NULL); break;
			case ROOK: new Rook(BLACK, *it, pNr, NULL); break;
			case QUEEN: new Queen(BLACK, *it, pNr, NULL); break;
			case KING: m_pKing[BLACK] = new King(BLACK, *it, pNr, NULL); break;
			}
			pNr--;
		}
	}

	//m_pKing[WHITE] = m_pieceList[15]; 
	//m_pKing[BLACK] = m_pieceList[31];
}

bool LearnObject::LoadTrainingSet(string filename, int filter)
{
fstream infile; 
bool toPlay, first = true;
int count = 0, f, maxValue = 0, value;
char c;
PositionSetIt start, end, it;

	infile.open(filename.c_str(), ios::in | ios::binary);

	if ( infile.good() )
	{
		m_currentSet = m_setList;
		m_currentSet->clear();
		m_currentSet->reserve(1500000); // test: quite a lot of memory is allocated
		infile.get(c);

		while ( c != NO_SQUARE && infile.good() )
		{
			INTVECTOR pieces[12];

			while ( c != NO_SQUARE && infile.good() )
			{
				f = c;
				infile.get(c);
				if ( c < 12 ) pieces[c].push_back(f);
				infile.get(c);
			} 

			infile.get(c);
			value = c + 128; 
			infile.get(c);
			toPlay = (c == WHITE ? WHITE : BLACK);

			if ( infile.good() ) 
			{
				if ( abs(value) > maxValue && value != 255 ) maxValue = abs(value);
				if ( FilterPosition(pieces, value, toPlay, filter) ) 
				{
					LoadPosition(pieces, toPlay);
					
					count++;
					if ( count % 10000 == 0 ) 
					{
						cout << ".";
						cout.flush();
					}
					if ( first )
					{
						cout << "The following position type has been detected: "
							 << m_evaluator.SetPositionType(AUTO_CHOOSE_POSITION_TYPE, true) 
							 << "\n"; 
						cout << "Loading training set: ";
						first = false;
					}

					AddCurrentPositionToSet(value, toPlay);
				}
			}

			infile.get(c);
		} 
		cout << " done.\n\n";

		// ----- transform position values to suit the learning function ------
		start = m_currentSet->begin(); 
		end = m_currentSet->end();
		for( it = start; it != end; it++ )
		{
			if ( it->value == 255 ) // position will lead to a draw
			{
				if ( it->tpPieces[0].atomics & BIT_0 ) // advantage for the side which will
					it->value = -ENDGAME_DRAW_VALUE;   // make the next move
				else if ( it->tpPieces[0].atomics & BIT_1 ) // disadvantage ...
					it->value = ENDGAME_DRAW_VALUE;
				else
					it->value = 0;
			}
			else
			{
				if ( it->value % 2 == 1 ) 
					it->value = maxValue + ENDGAME_VALUE_OFFSET - it->value;
				else 
					it->value = -(maxValue + ENDGAME_VALUE_OFFSET - it->value);
			}
		}
	}
	else
	{
		cout << "Error: " << filename << " not found!\n";
		return false;
	}

	infile.close();

	if ( m_currentSet->size() == 0 ) 
	{
		cout << "Error: Training set " << filename << " is empty or invalid!\n";
		return false;
	}
	else return true;
}

void LearnObject::WeightFitting()
{
double error, oldError = 0;
	m_evaluator.InitWeights();
	error = m_evaluator.Gstep(*m_currentSet);
	cout << "Weight fitting ";
	do 
	{
		oldError = error;
		error = m_evaluator.CGstep(*m_currentSet, false);
	} while ( oldError - error > 0.1 );
	cout << " done.\n";
	m_evaluator.SaveData();
	cout << "\n";
	cout << "Hit percentage: " << m_evaluator.GetHitPercentage(*m_currentSet) << "%\n\n";
}

void LearnObject::GenerateConfigurations(int maxA)
{
int count, temp;
unsigned int not_msac;

	for ( int piece = 0; piece < 6; piece++ )
	{
		// remove all the old configurations of this piece
		m_evaluator.RemoveConfigurations(piece);
		count = 0;
		not_msac = 0;

		// build a "tree" with all valid configurations

		// every configuration contains a relative jump address
		// for a faster evaluation, jump can be used to skip some of the
		// following (more specific) configurations, if a configuration is inactive

		// not_msac (not most specific active configuration) is another trick to
		// speed up the evaluation of the active configurations
		// if a configuration is a most specific active configuration, jump can again
		// be used to skip some configurations

		// maxA is used to limit the number of possible configurations
		// the maximum of atomic features of a configuration is limited to maxA + 1

		for ( int bit = 0; bit < 32; bit++ )
		{
			temp = GeneratePieceConfigurations(piece, BIT_0<<bit, bit+1, maxA, 0);//not_msac);
			
			if ( temp > 0 )
			{
				count += temp;
				not_msac |= BIT_0 << bit;
			}
		}
		cout << "Configuration count for piece " << piece << ": " << count <<"\n";
	}
	cout << "\n";
}

int LearnObject::GeneratePieceConfigurations(int piece, unsigned int bitcode, 
				 int nextbit, int maxA, unsigned int not_msac)
{
int count = 0, temp;

	// a possible configuration is a valid configuration as soon as it appears
    // in a several number of chess positions

	if ( !IsValidConfiguration(piece, bitcode, 15, 1000) ) return count;

	// the configuration must be stored before its more specific configurations
int cnr = m_evaluator.AddConfiguration(piece);
	m_evaluator.GetConfiguration(piece, cnr).bitcode = bitcode;
	count++;

	if ( maxA > 0 )
	{
		for ( int bit = nextbit; bit < 32; bit++ )
		{
			temp = GeneratePieceConfigurations(piece, bitcode|(BIT_0 << bit), 
				                               bit+1, maxA-1, 0);//not_msac);
			if ( temp > 0 )
			{
				count += temp;
				// the bit "bit" gave a more specific configuration -> update not_msac
				not_msac |= BIT_0 << bit;
			}
		}
	}
	else 
	{
		// there is no more specific active configuration
		not_msac = 0;
	}

Configuration& c = m_evaluator.GetConfiguration(piece, cnr);
	// count is the number of more specific active configurations followed by this configuration
    c.jump = count;
	c.not_msac = not_msac;
	return count;
}

bool LearnObject::IsValidConfiguration(int piece, unsigned int bitcode, int minCt, int maxIt)
{
PiecePropertiesSetIt estart, eend, eit;
PositionSetIt pit;
int count = 0, pCount;
unsigned int seed = 5, randomNumber;

	if ( m_currentSet->size() == 0 ) return false;

	for ( int i = 0; i < maxIt; i++ )
	{
		pCount = 0;
		seed = (seed * 239) & 65535;
		randomNumber = seed;
		seed = (seed * 239) & 65535;
		randomNumber = ((randomNumber << 16) | seed) % m_currentSet->size(); 
		pit = &(*m_currentSet)[randomNumber];

		estart = pit->tpPieces.begin(); eend = pit->tpPieces.end();
		for( eit = estart; eit != eend; eit++ )
			if ( eit->type == piece && (eit->atomics & bitcode) == bitcode ) pCount++;

		estart = pit->iePieces.begin(); eend = pit->iePieces.end();
		for( eit = estart; eit != eend; eit++ )
			if ( eit->type == piece && (eit->atomics & bitcode) == bitcode ) pCount--;

		count += abs(pCount);
		if ( count >= minCt ) return true;
	}

	return false;
}

void LearnObject::WDLStatistics()
{
int dbWin[3] = {0,0,0}, dbDraw[3] = {0,0,0}, dbLoss[3] = {0,0,0};
PositionSetIt pit, pend;
float value, size;

	pend = m_currentSet->end();
	for	( pit = m_currentSet->begin(); pit != pend; pit++ )
	{
		value = m_evaluator.GetExampleValue(*pit);

		if ( pit->value == -ENDGAME_DRAW_VALUE 
			 && pit->tpPieces[0].atomics & BIT_0 ) // draw with a material advantage
		{
			if ( value >= 0 ) dbDraw[0]++;
			else dbDraw[1]++;
		}
		else if ( pit->value == ENDGAME_DRAW_VALUE 
			      && pit->tpPieces[0].atomics & BIT_1 ) // draw with a disadvantage
		{
			if ( value <= 0 ) dbDraw[2]++;
			else dbDraw[1]++;
		}
		else if ( pit->value == 0 ) // draw with equal material
		{
			if ( value < -ENDGAME_VALUE_OFFSET ) dbDraw[0]++;
			else if ( value > ENDGAME_VALUE_OFFSET ) dbDraw[2]++;
			else dbDraw[1]++;
		}
		else if ( pit->value > 0 ) // winner
		{
			if ( value < 0 && pit->tpPieces[0].atomics & BIT_0 ) dbWin[1]++;
			else if ( value > 0 ) dbWin[0]++;
			else if ( value <= 0 ) dbWin[2]++;
		}
		else if ( pit->value <= 0 ) // loser
		{
			if ( value > 0 && pit->tpPieces[0].atomics & BIT_1 ) dbLoss[1]++;
			else if ( value < 0 ) dbLoss[2]++;
			else if ( value >= 0 ) dbLoss[0]++;
		}
	}

	m_evaluator.CGstep(*m_currentSet, true); // for statistics only

	size = ((float) m_currentSet->size()) * 0.01;
	if ( size == 0 ) return;
	cout.precision(3);

	cout << "\tGenerated\n";
	cout << "DB\tWin\tDraw\tLoss\n";
	cout << "Win\t" << dbWin[0] <<"\t" << dbWin[1] << "\t" << dbWin[2] << "\n";
	cout << "\t" << dbWin[0] / size  <<"%\t" 
		 << dbWin[1] / size << "%\t" << dbWin[2] / size << "%\n";
	cout << "Draw\t" << dbDraw[0] <<"\t" << dbDraw[1] << "\t" << dbDraw[2] << "\n";
	cout << "\t" << dbDraw[0] / size  <<"%\t" 
		 << dbDraw[1] / size << "%\t" << dbDraw[2] / size << "%\n";
	cout << "Loss\t" << dbLoss[0] <<"\t" << dbLoss[1] << "\t" << dbLoss[2] << "\n";
	cout << "\t" << dbLoss[0] / size  <<"%\t" 
		 << dbLoss[1] / size << "%\t" << dbLoss[2] / size << "%\n\n";
}

void LearnObject::HalfmoveStatistics(string filename, int filter, int sampling)
{
fstream infile; 
bool toPlay, first = true, isOk = true;
int count = 0, trialCount = 0, eSuccessCount = 0, rSuccessCount = 0, f, value;
char c;

	infile.open(filename.c_str(), ios::in | ios::binary);

	if ( infile.good() )
	{
		infile.get(c);

		while ( c != NO_SQUARE && infile.good() && isOk )
		{
			INTVECTOR pieces[12];

			while ( c != NO_SQUARE && infile.good() )
			{
				f = c;
				infile.get(c);
				if ( c < 12 ) pieces[c].push_back(f);
				infile.get(c);
			} 

			infile.get(c);
			value = c + 128; 
			infile.get(c);
			toPlay = (c == WHITE ? WHITE : BLACK);

			if ( infile.good() ) 
			{
				count++;
				if ( count % 10000 == 0 && !first ) 
				{
					cout << ".";
					cout.flush();
				}

				if ( FilterPosition(pieces, value, toPlay, filter) &&
					 count % sampling == 0 ) 
				{
					LoadPosition(pieces, toPlay);

					if ( first )
					{
						cout << "Doing halfmove statistics: ";
						m_evaluator.SetPositionType(AUTO_CHOOSE_POSITION_TYPE, true);
						first = false;
					}

					trialCount++;
					isOk =  HalfmoveSearch(eSuccessCount, rSuccessCount, value);
				}
			}

			infile.get(c);
		} 
		cout << " done.\n\n";
		if ( trialCount > 0 )
		{
			cout << "Halfmove hit percentage with evaluation function: " 
				 << 100.0 * (float) eSuccessCount / (float) trialCount << "%\n";
			cout << "Halfmove hit percentage without evaluation function: " 
				 << 100.0 * (float) rSuccessCount / (float) trialCount << "%\n\n";
		}
	}
	else
	{
		cout << "Error: " << filename << " not found!\n";
		return;
	}

	infile.close();
}

bool LearnObject::HalfmoveSearch(int &eSuccessCount, int &rSuccessCount, int value)
{
	NodeStack nstack(m_moveStack);
	m_currentStack = &nstack;
	Piece* p = m_pKing[m_move.toPlay];
	int tempValue, bestValue = IMPOSSIBLE_VALUE;
	bool isOk = true;

	do
	{
		p->generateMoves();
		p = p->next;
	} while ( p->type != KING );

	// statistics with evaluation function
	Move* m = m_moveStack;
	Move* best = NULL;
	Move* t = nstack.GetTail();

	while ( m < t )
	{
		m_move = *m;
		tempValue = pushMove();
		if ( !m_pKing[!m_move.toPlay]->isAttacked() ) 
		{
			// NegaScout will be happy with some "-" signs 
			tempValue = -m_evaluator.GetPositionValue(-tempValue, -tempValue, -tempValue);
			if ( tempValue > bestValue )
			{
				bestValue = tempValue;
				best = m;
			}
		}
		popMove();
		m++;
	}

	if ( best != NULL )
	{
		m_move = *best;
		pushMove();

		isOk = GetPositionValueFromDB(bestValue);
		
		if ( isOk && ((value >= 248 && bestValue >= 248) || value - 1 == bestValue) )
			eSuccessCount++;

		popMove();
	}
	else if ( isOk )
	{
		eSuccessCount++; // checkmate or draw
	}

	// statistics without evaluation function
	// a valid move is chosen at random
	m = m_moveStack;
	best = NULL;

	int rm = rand() % nstack.Size();
	int nrOfTrials = 0;

	while ( best == NULL && nrOfTrials < 200 )
	{
		m_move = m[rm];
		pushMove();

		if ( !m_pKing[!m_move.toPlay]->isAttacked() ) 
		{
			best = &(m[rm]);	
		}
		else // illegal move, let's choose another move
		{
			nrOfTrials++;
			rm = rand() % nstack.Size();
		}
		popMove();
	}

	if ( best != NULL )
	{
		m_move = *best;
		pushMove();

		isOk = GetPositionValueFromDB(bestValue);
		
		if ( isOk && ((value >= 248 && bestValue >= 248) || value - 1 == bestValue) )
			rSuccessCount++;

		popMove();
	}
	else if ( isOk )
	{
		rSuccessCount++; // checkmate or draw
	}

	return isOk;
}

bool LearnObject::GetPositionValueFromDB(int &value)
{
	return false;

	/*string str;
	char c;
	int f;
	unsigned int pos;
	FENTag tag;
	TCPConnection sdb;

	if ( !sdb.connect("129.132.16.3", 15901) )
	{
		cout << "\nError: Could not connect to server!\n";
		return false;
	}

	int b[64];

	for ( int i = 0; i < 64; i++ )
		b[i] = m_board[i]->hashtype;

	tag.SetTag(b, m_move.toPlay, NULL, 0, 1, m_epSquare);
	str = "chs_ps_\"" + tag.GetTag() + "\"\n";

	if ( !sdb.write(str) ) 
	{
		cout << "\nError: Could not write to server!\n";
		return false;
	}

	if ( !sdb.readline(str) )
	{
		cout << "\nError: Could not read from server!\n";
		return false;
	}

	value = 0;
	for ( pos = 7; pos < str.size(); pos++ )
	{
		c = str[pos];

		if ( c < '0' || c > '9' ) 
		{
			if ( c == '=' )
			{
				value = 255;
				return true;
			}
			else if ( pos > 7 ) // at least one digit has been read 
			{
				return true;
			}
			else // no digit has been read
			{
				cout << "\nError: Got confusing string from server!\n";
				return false;
			}
		}

		f = c - '0';
		value = 10 * value + f;
	}

	if ( pos == 7 )
	{
		cout << "\nError: Got confusing string from server!\n";
			return false;
	}

	return true;*/
}

/*void LearnObject::ProcessPGNFile(PGNFile* f)
{
	PGNGame* g;
	Piece* p;
	int nr = 0, result, value, lastValue;
	int from_r, from_c, to_r, to_c, piece, p_piece;

	for ( int i = 0; i < MAX_POSITIONTYPES; i++ ) 
	{
		m_setList[i].clear();
		m_setList[i].reserve(1000);
	}

	cout << "Processing pgn games: ";

	// load all chess games from the file
	g = f->GetGame(nr);
	while ( g != NULL )
	{
		// get the final result of the chess game, skip games with unknown results
		result = g->GetResult();
		while ( result == PGN_RESULT_UNKNOWN )
		{
			nr++;
			g = f->GetGame(nr);
			if ( g == NULL ) return;
			result = g->GetResult();
		}

		// load the initial position of a valid game
		LoadPosition(g->pieces, g->activeColor);

		// load the game
		bool valid = true;
		while ( g->GetNextMove(from_r, from_c, to_r, to_c, piece, p_piece) && valid )
		{
			valid = DoPGNMove(from_r, from_c, to_r, to_c, piece, p_piece);
		}

		// add up the piece values of the final position
		lastValue = 0;
		p = m_pKing[WHITE]; // count white piece values positive
		do
		{
			lastValue += p->value;
			p = p->next;
		} while ( p->type != KING );

		p = m_pKing[BLACK]; // count black piece values negative
		do
		{
			lastValue -= p->value;
			p = p->next;
		} while ( p->type != KING );

		// add a bonus for the winning side
		if ( result == PGN_WHITE_WINS ) lastValue += 5 * Q_PAWN;
		else if ( result == PGN_BLACK_WINS ) lastValue -= 5 * Q_PAWN;

		// dont take games that are too short (< 10 halfmoves)
		// or games that could not be properly loaded (valid)
		if ( m_gameHistory->GetCount() > 10 && valid ) 
		{ 
			while ( m_gameHistory->GetCount() > 0 )
			{
				// add up the piece values of the current position
				value = 0;
				p = m_pKing[WHITE];
				do
				{
					value += p->value;
					p = p->next;
				} while ( p->type != KING );

				p = m_pKing[BLACK];
				do
				{
					value -= p->value;
					p = p->next;
				} while ( p->type != KING );

				// choose the right position set for the current position
				m_currentSet = 
					&m_setList[m_evaluator.SetPositionType(AUTO_CHOOSE_POSITION_TYPE, false)];

				// the value of the current position depends on the 
				// piece sum of the current and the following positions of the game
				lastValue = 0.1 * value + 0.9 * lastValue;
				if ( lastValue >= 0 )
					value = 0.08 * Q_PAWN * log(lastValue / (double) Q_PAWN + 1.0);
				else
					value = - 0.08 * Q_PAWN * log(-lastValue / (double) Q_PAWN + 1.0);

				// add the position to the current set
				if ( m_move.toPlay == WHITE ) 
					AddCurrentPositionToSet(value, m_move.toPlay);
				else
					AddCurrentPositionToSet(-value, m_move.toPlay);
			
				popMove(); // undo one halfmove
			}
		}

		cout << ".";
		cout.flush();
		nr++;
		g = f->GetGame(nr);
	}
	cout << " done.\n\n";
}*/

void LearnObject::ProcessPGNFile(PGNFile* f)
{
	PGNGame* g;
	Piece* p;
	int nr = 0, nrOfHalfmoves, result, value;
	int from_r, from_c, to_r, to_c, piece, p_piece;

	for ( int i = 0; i < MAX_POSITIONTYPES; i++ ) 
	{
		m_setList[i].clear();
		m_setList[i].reserve(1000);
	}

	cout << "Processing pgn games: ";

	g = f->GetGame(nr);
	while ( g != NULL )
	{
		result = g->GetResult();
		while ( result == PGN_RESULT_UNKNOWN )
		{
			nr++;
			g = f->GetGame(nr);
			if ( g == NULL ) return;
			result = g->GetResult();
		}

		LoadPosition(g->pieces, g->activeColor);

		bool valid = true;
		int qWhite = 0, qBlack = 0;
		float qF;

		while ( g->GetNextMove(from_r, from_c, to_r, to_c, piece, p_piece) && valid )
		{
			valid = DoPGNMove(from_r, from_c, to_r, to_c, piece, p_piece);

			p = m_pKing[WHITE];
			do
			{
				qWhite += p->value;
				p = p->next;
			} while ( p->type != KING );

			p = m_pKing[BLACK];
			do
			{
				qBlack += p->value;
				p = p->next;
			} while ( p->type != KING );
		}

		if ( qWhite > qBlack ) qF = ((qWhite / (float) qBlack) - 0.95) * 5;
		else qF = ((qBlack / (float) qWhite) - 0.95) * 5;

		nrOfHalfmoves = m_gameHistory->GetCount();

		if ( nrOfHalfmoves > 10 && valid ) // security
		{
			int qDiff, qLastDiff[4] = {0,0,0,0}; 

			while ( m_gameHistory->GetCount() > 0 )
			{
				qDiff = 0;
				p = m_pKing[WHITE];
				do
				{
					qDiff += p->value;
					p = p->next;
				} while ( p->type != KING );

				p = m_pKing[BLACK];
				do
				{
					qDiff -= p->value;
					p = p->next;
				} while ( p->type != KING );

				qDiff = abs(qDiff);

				if ( qDiff < 10 * Q_PAWN && 
					(abs(qDiff - qLastDiff[1]) < Q_PAWN || abs(qDiff - qLastDiff[3]) < Q_PAWN) )
				{
					m_currentSet = 
						&m_setList[m_evaluator.SetPositionType(AUTO_CHOOSE_POSITION_TYPE, false)];
					value = (int) (150 * qF * result * (m_gameHistory->GetCount() / (float) nrOfHalfmoves));

					if ( m_move.toPlay == WHITE ) 
						AddCurrentPositionToSet(value, m_move.toPlay);
					else
						AddCurrentPositionToSet(-value, m_move.toPlay);
				}

				qLastDiff[3] = qLastDiff[2];
				qLastDiff[2] = qLastDiff[1];
				qLastDiff[1] = qLastDiff[0];
				qLastDiff[0] = qDiff;
			
				popMove(); // first position will not be added to the training set
			}
		}

		cout << ".";
		cout.flush();
		nr++;
		g = f->GetGame(nr);
	}
	cout << " done.\n\n";
}

void LearnObject::PGNWeightFitting()
{
	clock_t start, stop; // just to get an idea how much time has been used

	// special endgames are trained using databases!
	for ( int i = ENDGAME_1; i < MAX_POSITIONTYPES; i++ ) 
	{
		start = clock();

		cout << "Fitting weights of positiontype "<< i << ".\n";
		m_currentSet = &m_setList[i];
		m_evaluator.SetPositionType(i, true);
		WeightFitting();
		m_evaluator.CGstep(*m_currentSet, true); // for statistics only

		stop = clock(); 
		double duration = (double)(stop - start) / CLOCKS_PER_SEC;
		cout << "Time: " << duration << " seconds\n\n";
	}
}

void LearnObject::PGNStatistics()
{
	// not yet available
}

void LearnObject::GeneratePGNConfigurations()
{
	// special endgames are trained using databases!
	for ( int i = ENDGAME_1; i < MAX_POSITIONTYPES; i++ ) 
	{
		cout << "Number of positions of type "<< i << ": " << m_setList[i].size() << "\n";
		m_currentSet = &m_setList[i];
		m_evaluator.SetPositionType(i, true);
		GenerateConfigurations(2);
	}
}

void LearnObject::OptimizeConfigurationSet()
{
	m_evaluator.OptimizeConfigurationSet();
	m_evaluator.SaveData();
}
