/***************************************************************************
                   pgnfile.h  -  handling of pgn games
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// disable warning C4786: symbol greater than 255 character,
#ifndef GNU_VERSION
#pragma warning(disable: 4786) // Visual C++ stuff
#endif

#include <vector>
#include <string>

using namespace std ;

typedef vector<string> STRINGVECTOR;
typedef vector<int> INTVECTOR;

#define PGN_PAWN 0
#define PGN_KNIGHT 1
#define PGN_BISHOP 2
#define PGN_ROOK 3
#define PGN_QUEEN 4
#define PGN_KING 5
#define PGN_NO_PIECE 6

// valid columns and ranks have numbers between 0 and 7
#define PGN_NOT_SPECIFIED 8

// squares have numbers between 0 and 63 (square = column + 8 * rank)
#define PGN_NO_SQUARE 64

// numbers of the pieces
#define PGN_B_PAWN 0
#define PGN_B_ROOK 1
#define PGN_B_BISHOP 2
#define PGN_B_KNIGHT 3
#define PGN_B_QUEEN 4
#define PGN_B_KING 5
#define PGN_W_PAWN 6
#define PGN_W_ROOK 7
#define PGN_W_BISHOP 8
#define PGN_W_KNIGHT 9
#define PGN_W_QUEEN 10
#define PGN_W_KING 11
#define PGN_EMPTY_SQUARE 12

// castling availability (white/black, kingside/queenside)
#define PGN_W_KS_CASTLING 0
#define PGN_W_QS_CASTLING 1
#define PGN_B_KS_CASTLING 2
#define PGN_B_QS_CASTLING 3

// piece colours
#define PGN_BLACK 0
#define PGN_WHITE 1

// game results
#define PGN_WHITE_WINS        1
#define PGN_BLACK_WINS       -1
#define PGN_DRAW			  0
#define PGN_RESULT_UNKNOWN    2

class FENTag
{
public:
	FENTag() { tag = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"; };
	string SetTag(int* board, bool toPlay = PGN_WHITE, 
				  bool* castlingAvailability = NULL,
				  int halfmoveClock = 0, int fullmoveNumber = 1, 
				  int epSquare = PGN_NO_SQUARE );
	string GetTag() { return tag;};
private:
	string tag;
};

class PGNGame
{
public:
	PGNGame() { currentElement = 0;}
	STRINGVECTOR tags;
	STRINGVECTOR moveText;

	// initial settings:
	INTVECTOR pieces[12]; 
	bool activeColor;
	bool castlingAvailability[4];
	int epSquare;
	int halfmoveClock;
	int fullmoveNumber;
	
	bool GetNextMove( int& from_r, int& from_c, int& to_r, 
					  int& to_c, int& piece, int& p_piece );
	bool UndoLastMove();
	bool ParseFenTag(string str);
	int GetResult();
	void Reset() { currentElement = 0; };
private:
	int currentElement;
};

typedef vector<PGNGame> PGNGAMEVECTOR;

class PGNFile
{
public:
	PGNFile() {};
	~PGNFile() {};
	bool LoadFile(string name);
	PGNGame* GetGame(int nr);
private:
	PGNGAMEVECTOR gameList;
};


