/***************************************************************************
        pieces.h  -  all the pieces with their special behaviour
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//----------------------------------------------------------------------------
//		BASECLASS
//----------------------------------------------------------------------------

class Piece:public EngineObject
{
protected:
	void init(int _color, int _square, int _number, Piece* copy);

	int promotion()
	{
		m_move.o_value += 10001;
		m_move.special = NEW_QUEEN;
		m_currentStack->AddElement(&m_move);
		m_move.o_value--;
		m_move.special = NEW_KNIGHT;
		m_currentStack->AddElement(&m_move);
		//m_move.special = NEW_BISHOP; //test: cp count += 2; -> count += 4;
		//m_currentStack->AddElement(&m_move);
		//m_move.special = NEW_ROOK;
		//m_currentStack->AddElement(&m_move);
		m_move.o_value = 0;
		return 2; //Number of elements added
	};

	void diagonal();
	void straight();
public:
	int hashtype; // information about type and color of the piece
	const int* possibleMoves; // pointer to array containing possible moves (relative 
							  // distance between two squares)

	Piece() {};
	Piece(int _color, int _square, int _number, Piece* copy);

	bool onKingsLine(); // if a piece which is "onKingsLine" is moved, the own king
						// could get attacked
	virtual void verifyMove(); // verify the current move m_move with given sq1 and sq2
							   // and set the additional move information of m_move
	bool isAttacked(); // verify whether the piece is attacked by the opponent
	virtual int generateMoves(); // generate all possible moves and append them to a list
	virtual void generateXMoves(); // generate all capture moves and append them to a list

	// pieces of the same color build a doubly linked list
	Piece* previous; // previous piece in the list
	Piece* next; // next piece in the list

	int drawValue; // the draw value of the piece (used to calculate draw configurations)
	int bonus[64]; // information for move ordering
	int number; // number of the piece

	int square; // placement of the piece on the board
	int moveCount; // number of moves the piece has done
	int type; // type of the piece (see def_inc.h for possible types)
	int color; // color of the piece
	int value; // piece value (cp. def_inc.h)

	unsigned int atomics; //atomic features of the piece
	virtual void evaluateAtomics() {}; // calculate atomics for the given position
};

//----------------------------------------------------------------------------
//		PAWN, white and black
//----------------------------------------------------------------------------

class WPawn:public Piece
{
public:
	virtual void generateXMoves();
	virtual int generateMoves();
	WPawn(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove();
	virtual void evaluateAtomics();
};

class BPawn:public Piece
{
public:
	virtual void generateXMoves();
	virtual int generateMoves();
	BPawn(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove();
	virtual void evaluateAtomics();
};

//----------------------------------------------------------------------------
//		KNIGHT
//----------------------------------------------------------------------------

class Knight:public Piece
{
public:
	virtual int generateMoves();
	virtual void generateXMoves();
	Knight(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove();
	virtual void evaluateAtomics();
};

//----------------------------------------------------------------------------
//		BISHOP
//----------------------------------------------------------------------------

class Bishop:public Piece
{
public:
	Bishop(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove() { m_move.special = NOTHING; diagonal();};
	virtual void evaluateAtomics();
};

//----------------------------------------------------------------------------
//		ROOK
//----------------------------------------------------------------------------

class Rook:public Piece
{
public:
	Rook(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove() {
		m_move.special = NOTHING;
		straight();};
	virtual void evaluateAtomics();
};

//----------------------------------------------------------------------------
//		QUEEN
//----------------------------------------------------------------------------

class Queen:public Piece
{
public:
	Queen(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove();
	virtual void evaluateAtomics();
};

//----------------------------------------------------------------------------
//		KING
//----------------------------------------------------------------------------

class King:public Piece
{
public:
	virtual int generateMoves();
	virtual void generateXMoves();
	King(int _color, int _square, int _number, Piece* copy);
	virtual void verifyMove();
	virtual void evaluateAtomics();
	bool m_bCastled;
};

