/***************************************************************************
   winboardinterface.cpp  -  connection between engine and WinBoard
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GNU_VERSION
#include <windows.h>
#endif

#include <ctime>
#include <iostream>
#include <fstream>
using namespace std;

#include "def_inc.h"
#include "engineobject.h"
#include "listclasses.h"
#include "pieces.h"
#include "hashtable.h"
#include "evaluation.h"
#include "engine.h"
#include "winboardinterface.h"


// communication between WinBoard and ChessterfieldCL is multithreaded,
// win32 threads are used for the windows platform
DWORD m_inputThreadID;
DWORD m_promptThreadID;
HANDLE m_hEventInputReady = CreateEvent(NULL,TRUE,FALSE,NULL);
string mt_input;
bool bShowPrompt = true;

// ChessterfieldCL shows some information if it is started without WinBoard
DWORD WINAPI ShowPromptThread(LPVOID pParam)
{
	Sleep(100); // give WinBoard the chance to send the xboard command

	if ( bShowPrompt )
	{
		cout << endl;
		cout << "---------------------------------------------------------\n";
		cout << "            ChessterfieldCL - WinBoard engine\n";
		cout << "---------------------------------------------------------\n";
		cout << "Author: Matthias Luescher (m.luescher@datacomm.ch)\n";
		cout << "Copyright: Matthias Luescher\n";
		cout << "Web: http://www.luescher-online.com/computerchess.html\n";
		cout << "Version: i5a (unofficial release)\n\n";
		cout << "Tip: Use Tim Mann's WinBoard as a user interface\n";
		cout << "Initstring: ChessterfieldCl /fd=\\\"C:\\\\SomeDirectory\\\"\n";;
		cout << "---------------------------------------------------------\n\n";
		cout << "Type 'quit' to exit the program or enter your move!" << endl;
	}

	return 0;
}

DWORD WINAPI InputThread(LPVOID pParam)
{
	bool quit = false;
	WinBoardInterface* wbi = (WinBoardInterface*) pParam;

	while ( !quit )
	{
		cin >> mt_input;

		if ( mt_input == "post" )
			wbi->engine->m_bPostThinking = true;
		else if ( mt_input == "nopost" )
			wbi->engine->m_bPostThinking = false;
		else if ( mt_input == "." && !wbi->bEditMode )
			wbi->engine->m_bPostAnalysis = true;
		else if ( mt_input == "?" )
			wbi->engine->m_bStopEngine = true;
		else if ( mt_input == "hard" )
			wbi->bPermanentBrain = true;
		else if ( mt_input == "easy" )
			wbi->bPermanentBrain = false;
		else if ( mt_input == "random" ) 
		{
			// toggle "random mode"
			// not implemented
		}
		else
			quit = true;
	}

	wbi->bUserInputReady = true;
	wbi->engine->m_bStopEngine = true;
	SetEvent(m_hEventInputReady);

	return 0;
}

string WinBoardInterface::ListenToWinBoard(bool blocking)
{
static bool bThreadLock = false;

	if ( !bThreadLock )
	{
		CreateThread(NULL, 0, InputThread, this, 0, &m_inputThreadID);
		bThreadLock = true;
	}

	// if blocking == true wait until there is new user input, otherwise
	// continue with the main thread
	if ( blocking )
	{
		WaitForSingleObject(m_hEventInputReady, INFINITE);
		ResetEvent(m_hEventInputReady);
		bThreadLock = false;
		bUserInputReady = false;
		return mt_input;
	}

	return "";
}


WinBoardInterface::WinBoardInterface()
{
	engine = new Engine(NULL, WHITE);
	bComputerBlack = true;
	bComputerWhite = false;
	bPermanentBrain = false;
	bEditMode = false;
	remainingTime[0] = 300000;
	remainingTime[1] = 300000;

	movesPerTimecontrol = 0;
	baseTime = 300000;
	incrementalTime = 0;
	searchMode = TIME_LIMIT;
	searchDepth = MAX_SEARCH_DEPTH;

	bUserInputReady = false;

	CreateThread(NULL, 0, ShowPromptThread, NULL, 0, &m_promptThreadID);
}

WinBoardInterface::~WinBoardInterface()
{
	delete engine;
}


long WinBoardInterface::GetIncrementalTimeFromString(string str)
{
	int multiplicator = 1, seconds = 0, n, i;

	for (i = str.size() - 1; i >= 0 ; i--)
	{
		n = str[i] - '0';
		if ( n >= 0 && n <= 9 )
		{
			seconds += n * multiplicator;
			multiplicator *= 10;
		}
		else 
		{
			multiplicator = 60;
		}
	}
	return 1000 * seconds; // incremental time in milliseconds
}

long WinBoardInterface::GetBaseTimeFromString(string str)
{
	int multiplicator = 60, seconds = 0, n, i;

	for (i = str.size() - 1; i >= 0 ; i--)
	{
		n = str[i] - '0';
		if ( n >= 0 && n <= 9 )
		{
			seconds += n * multiplicator;
			multiplicator *= 10;
		}
		else // minutes -> seconds
		{
			seconds /= 60;
			multiplicator = 60;
		}
	}
	return 1000 * seconds; // base time unit is milliseconds
}

bool WinBoardInterface::IsValidSetup(int* board)
{
int testsum[] = { 8, 2, 2, 2, 1, 1,
				  8, 2, 2, 2, 1, 1, 64 };

	for ( int square = 0; square < 64; square++ )
	{
		if ( ( square / 8 == 0 || square / 8 == 7 ) && 
			 ( board[square] == W_PAWN || board[square] == B_PAWN ) )
		{
			cout << "telluser Error: Illegal pawn positions!" << endl;
			return false;
		}

		if ( testsum[board[square]] > 0 )
			testsum[board[square]]--;
		else
		{
			if ( board[square] == W_KING || board[square] == B_KING )
			{
				cout << "telluser Error: Wrong number of kings!" << endl;
				return false;
			}
			else
			{
				testsum[board[square] - board[square] % 6]--;
			}
		}
	}

	if ( testsum[W_PAWN] < 0 || testsum[B_PAWN] < 0 )
	{
		cout << "telluser Error: Wrong number of pieces!" << endl;
		return false;
	}

	if ( testsum[W_KING] != 0 || testsum[B_KING] != 0 )
	{
		cout << "telluser Error: Wrong number of kings!" << endl;
		return false;
	}

	return true;
}

bool WinBoardInterface::ReadWinBoardMove(string str, int& from, int& to, int& castlingType)
{
int x1,x2,y1,y2;
	x1 = str[0] - 'a'; y1 = str[1] - '1';
	x2 = str[2] - 'a'; y2 = str[3] - '1';

	if ( str.size() < 4 )
		return false;

	if ( x1 >= 0 && x1 < 8 && y1 >= 0 && y1 < 8 )
		from = 8 * y1 + x1;
	else 
		return false;

	if ( x2 >= 0 && x2 < 8 && y2 >= 0 && y2 < 8 )
		to = 8 * y2 + x2;
	else 
		return false;

	if ( str.size() > 4 )
	{
		switch( str[4] )
		{
		case 'q': castlingType = NEW_QUEEN; break;
		case 'r': castlingType = NEW_ROOK; break;
		case 'b': castlingType = NEW_BISHOP; break;
		case 'n': castlingType = NEW_KNIGHT; break;
		default: castlingType = NOTHING; break;
		}
	}
	else
		castlingType = NOTHING;

	return true;
}

void WinBoardInterface::CalculateSearchTime(long& timeSoftStop, long& timeHardStop)
{
	long remMoves, dt, latency = 1000;
	int moveNr = engine->m_gameHistory->GetCount();

	if ( incrementalTime == 0 && movesPerTimecontrol != 0 ) // conventional clock
	{
		remMoves = (movesPerTimecontrol-((moveNr/2)%movesPerTimecontrol));
		dt = remainingTime[engine->toPlay()] / remMoves;
		latency = 1000 * remMoves;
	}
	else // incremental clock
	{
		if ( incrementalTime != 0 )
		{
			if ( moveNr < 96 )
				dt = remainingTime[engine->toPlay()] / (50 - moveNr/2) + incrementalTime;
			else
				dt = remainingTime[engine->toPlay()] / 2;
		}
		else
		{
			if ( moveNr < 60 )
				dt = remainingTime[engine->toPlay()] / 40;
			else if ( moveNr < 100 )
				dt = remainingTime[engine->toPlay()] / 30;
			else 
				dt = remainingTime[engine->toPlay()] / 20;
		}
	}

	// the strategy is to calculate two different times:
	// timeHardStop: the engine must stop
	// timeSoftStop: the engine can stop if it has completed a certain
	//				 part of the search tree
	timeHardStop = (dt * 7)/3;
	timeSoftStop = (dt * 1)/3;

	if (timeHardStop > remainingTime[engine->toPlay()] - latency )
		timeHardStop = remainingTime[engine->toPlay()] - latency;

	long offset = GetTickCount();

	timeSoftStop += offset;
	timeHardStop += offset;
}

bool WinBoardInterface::DoEditMode() 
{
	bool color = WHITE; 
	int board[64];
	string str;

	for ( int square = 0; square < 64; square++ ) board[square] = EMPTY_SQUARE;

	do
	{
		str = ListenToWinBoard(true);

		if ( str == "#" )
		{
			for ( int square = 0; square < 64; square++ )
				board[square] = EMPTY_SQUARE;
		}
		else if ( str == "." )
		{
			if ( IsValidSetup(board) )
			{
				bool toPlay = engine->toPlay();
				delete engine;
				engine = new Engine(board, toPlay);
				bEditMode = false;
				return true;
			}
			else
			{
				cout << "Error (command not legal now): ." << endl;
			}
		}
		else if ( str == "c" )
		{
			color = !color;
		}
		else if ( str == "quit" ) {}
		else // try to read the next piece from the input
		{
			int x1,y1, square, piece;
			bool valid = true;
			x1 = str[1] - 'a'; y1 = str[2] - '1';

			if ( x1 >= 0 && x1 < 8 && y1 >= 0 && y1 < 8 )
				square = 8 * y1 + x1;
			else 
				valid = false;

			switch( str[0] )
			{
			case 'Q': piece = (color == WHITE ? W_QUEEN : B_QUEEN); break;
			case 'K': piece = (color == WHITE ? W_KING : B_KING); break;
			case 'R': piece = (color == WHITE ? W_ROOK : B_ROOK); break;
			case 'B': piece = (color == WHITE ? W_BISHOP : B_BISHOP); break;
			case 'N': piece = (color == WHITE ? W_KNIGHT : B_KNIGHT); break;
			case 'P': piece = (color == WHITE ? W_PAWN : B_PAWN); break;
			case 'x': piece = EMPTY_SQUARE; break;
			default: valid = false; break;
			}

			if ( valid )
				board[square] = piece;
			else
				cout << "Error (command not legal now): " << str << endl;
		}
	} while ( str != "quit" );

	return false;
}

bool WinBoardInterface::DoPlayMode()
{
	string input; 
	bool resetHash = true;
	int from, to, castlingType;

	while(true)
	{
		if ((engine->toPlay() == WHITE && bComputerWhite) || // Try to make a computer move
		    (engine->toPlay() == BLACK && bComputerBlack))
		{
			engine->m_bStopEngine = false;
			engine->m_bSoftStop = false;
			long timeSoftStop, timeHardStop;
			CalculateSearchTime(timeSoftStop, timeHardStop);

			ListenToWinBoard(false);

			if ( engine->MakeComputerMove(timeSoftStop, timeHardStop, 
										  searchDepth, searchMode, resetHash))
			{
				cout << "move " << engine->GetLastMove() << endl;

				resetHash = true;
				engine->m_bStopEngine = false;
				engine->m_bSoftStop = false;

				// pondering stuff ->
				if ( bPermanentBrain && !engine->m_bAnalyzeMode &&
					 !bUserInputReady )
				{
					string m = engine->GetPV1Move();
					if ( ReadWinBoardMove(m, from, to, castlingType))  
					{
						if ( engine->MakeUserMove(from, to, castlingType) )
						{
							if ( engine->MakeComputerMove(0, 0, 0, NO_LIMITS, resetHash) )
							{
								resetHash = false;
								engine->UndoMove();
							}
							engine->UndoMove();
						}
					}
				}
				// <- pondering stuff
			}
			
			switch ( engine->GetState() )
			{
			case CHECKMATE:
				if ( engine->toPlay() == BLACK ) cout << "1-0 {White mates}" << endl;
				else cout << "0-1 {Black mates}" << endl;
				break;
			case REMIS: cout << "1/2-1/2 {Remis}" << endl; break;
			case PATT: cout << "1/2-1/2 {Stalemate}" << endl; break;
			case REPETITION_RULE: cout << "1/2-1/2 {Draw by repetition}" << endl; break;
			}
		}
		
		input = ListenToWinBoard(true);

		if ( ReadWinBoardMove(input, from, to, castlingType)) // see whether the 
		{													  // input is a move
			if ( !engine->MakeUserMove(from, to, castlingType))
				cout << "Illegal move: " << input << endl;
		}
		else if ( input == "xboard" ) // switch to xboard mode
		{
			bShowPrompt = false;
			cout << endl;
		}
		else if ( input == "new" ) // start a new game
		{
			bComputerWhite = false;
			bComputerBlack = true;
			searchMode = TIME_LIMIT;
			searchDepth = MAX_SEARCH_DEPTH;

			delete engine;
			engine = new Engine(NULL, WHITE);	
		}
		else if ( input == "quit" ) // quit the program
		{
			return false;
		}
		else if ( input == "force" ) // switch the engine to "force mode"
		{
			bComputerBlack = false;
			bComputerWhite = false;
		}
		else if ( input == "level" ) // set some search options
		{
			string str;
			long MPS, BASE, INC;
			cin >> MPS;
			cin >> str;
			BASE = GetBaseTimeFromString(str);
			cin >> str;
			INC = GetIncrementalTimeFromString(str);
			if ( !cin.good() ) cin.clear();
			else 
			{
				movesPerTimecontrol = MPS;
				baseTime = BASE;
				incrementalTime = INC;
			}
		}
		else if ( input == "sd" ) // set a search depth
		{						  // and switch of the search time limit
			int depth;
			cin >> depth;
			if ( !cin.good() ) cin.clear();
			else 
			{
				searchMode = DEPTH_LIMIT;
				searchDepth = depth;
			}
		}
		else if ( input == "go" ) // set the engine to play the next move
		{
			if ( engine->toPlay() == WHITE )
			{
				bComputerWhite = true;
				bComputerBlack = false;
			}
			else
			{
				bComputerWhite = false;
				bComputerBlack = true;
			}
		}
		else if ( input == "time" ) 
		{
			long centisecs;
			cin >> centisecs;
			if ( !cin.good() ) 
			{
				cin.clear();
				centisecs = 10000;
			}

			if ( bComputerWhite ) SetRemainingTime(WHITE, 10 * centisecs);
			else SetRemainingTime(BLACK, 10 * centisecs);
		}
		else if ( input == "otim" )
		{
			long centisecs;
			cin >> centisecs;
			if ( !cin.good() ) 
			{
				cin.clear();
				centisecs = 10000;
			}

			if ( bComputerWhite ) SetRemainingTime(BLACK, 10 * centisecs);
			else SetRemainingTime(WHITE, 10 * centisecs);
		}
		else if ( input == "white" )
		{
			engine->setToPlay(WHITE);
			bComputerBlack = true;
			bComputerWhite = false;
		}
		else if ( input == "black" )
		{
			engine->setToPlay(BLACK);
			bComputerBlack = false;
			bComputerWhite = true;
		}
		else if ( input == "draw" )
		{
			// not implemented
			//if ( m_zug.value <= 0 )
			//	cout << "offer draw" << endl;
		}
		else if ( input == "edit" )
		{
			bEditMode = true;
			return true;
		}
		else if ( input == "undo" )
		{
			engine->UndoMove();
		}
		else if ( input == "remove" )
		{
			engine->UndoMove();
			engine->UndoMove();
		}
		else if ( input == "hint" ) // doesnt work while engine is pondering
		{
			string m = engine->GetPV1Move();
			if ( m != "") cout << "Hint: " << m << endl;
		}
		else if ( input == "analyze" )
		{
			engine->m_bAnalyzeMode = true;
			bComputerBlack = true;
			bComputerWhite = true;
		}
		else if ( input == "exit" ) // quit analyze mode
		{
			engine->m_bAnalyzeMode = false;
			bComputerBlack = false;
			bComputerWhite = false;
		}
		else
		{
			cout << "Error (unknown command): " << input << endl;
		}
	}
}

