/***************************************************************************
      winboardinterface.h  -  connection between engine and WinBoard
                             -------------------
    begin                : Sat Jun 3 2000
    copyright            : (C) 1996-2001 by Matthias Luescher
    email                : m.luescher@datacomm.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <string>
#include <iostream>

using namespace std;

class Engine;

// IMPORTANT HINT: WinBoardInterface is a class that is responsible for the
// communication between WinBoard and ChessterfieldCL. This class has been implemented
// according to the information given in the engine-intf.html file that comes with
// WinBoard

class WinBoardInterface  
{
public:
	WinBoardInterface();
	~WinBoardInterface();

	// default is play mode, to edit a position ChessterfieldCL
	// switches to the edit mode
	bool DoPlayMode();
	bool DoEditMode();

	// a pointer to the move generator
	Engine* engine;

	// members to store WinBoard settings
	bool bPermanentBrain;
	bool bEditMode;

	// inform ChessterfieldCL about new input from the interface
	bool bUserInputReady;
private:
	// helper functions to parse WinBoard commands:
	long GetIncrementalTimeFromString(string str);
	long GetBaseTimeFromString(string str);
	bool ReadWinBoardMove(string str, int& from, int& to, int& castlingType);

	// check wether the edited position is possible
	bool IsValidSetup(int* board);

	// time management functions
	void CalculateSearchTime(long& timeSoftStop, long& timeHardStop);
	void SetRemainingTime(bool color, long milliseconds) 
	{ 
		remainingTime[color] = milliseconds; 
	};

	// communication with WinBoard is multithreaded
	string ListenToWinBoard(bool blocking);

	int searchMode; // see def_inc.h for possible search modes

	// members to store WinBoard settings
	bool bComputerBlack;
	bool bComputerWhite;
	long remainingTime[2];
	long movesPerTimecontrol;
	long baseTime;
	long incrementalTime;
	int searchDepth;
};
